open Globals;;
open Typing;;
open Pattern;;
open Location;;

let machine symtable =
      match !mach with
        GOI | Question -> Gqnam.code symtable
      | GOInopt -> Nonoptgnam.code symtable
      | IrrevKriv(Big) -> Kall.code Knambis.machine symtable
      | IrrevValue(Big) -> Kall.code Kvalbis.machine symtable 
      | IrrevKriv(C) -> Kall.code Cknam.machine symtable
      | IrrevValue(C) -> Kall.code Ckval.machine symtable
      | IrrevKriv(GC) -> Kall.code Knam.machine symtable
      | IrrevValue(GC) -> Kall.code Kval.machine symtable
      | IrrevNeed(GC) -> Kall.code Knee.machine symtable
      | IrrevNeed(_) -> failwith "Internal error : the call-by-need machine \
                        cannot copy environments" 
;;

let anal_phrase lexbuf =
   try
       idcount := 0;
       let p=Parser.prog Lexer.main lexbuf in
       reset_type_vars();
       let ty = type_prog p in
       let p2 = optimizeprog p in
       if !verbose then print_prog stdout p2;
       (ty,p2)
   with
     Parsing.Parse_error ->
       output_location (Lexing.lexeme_start lexbuf,
                        Lexing.lexeme_end lexbuf);
       print_endline "Syntax error.";
       raise Bad_end
   | Lexer.Lex_error (errcode, pos1, pos2) ->
       (match errcode with
           Lexer.Unterminated_comment -> 
             output_file();
             print_endline "Lexical error. Unterminated comment."
         | Lexer.Illegal_character ->
             output_location (pos1, pos2);
             print_endline "Lexical error. Illegal character.");
       raise Bad_end
;;

let outname = ref "a.c";;
let exename = ref "";;

let compile filename =
   let inputc = open_in_bin filename in
   let lexbuf = Lexing.from_channel inputc in
   act_input := File(filename,inputc);
   numexpr := 0;
   globalsym := makegsym !glob;
   let filesym = ref !glob in
      (try
         while true do
            match anal_phrase lexbuf with
              (ty,Affect(x,e)) -> filesym := (x, ty, e, x) :: !filesym
            | (ty,Expr e) -> incr numexpr;
                             let x = "_expr_" ^ (string_of_int !numexpr) in
                                filesym := (x, ty, e, x) :: !filesym
         done
      with 
         End_of_file ->
            close_in inputc;
            oc := open_out !outname;
            (try
               machine !filesym;
               close_out !oc
            with
               x -> close_out !oc;
                    raise x)
       | x -> close_in inputc;
              raise x);
   if !exename <> "" then
   begin
      print_string "Compiling C file...";
      print_newline();
      if Sys.command(Config.gcc ^ " -o " ^ !exename ^ " " ^ !outname)<>0
      then begin
         print_endline "C compilation failed. This should never happen";
         raise Bad_end
      end
   end
;;

let secure compile filename =
   act_input := File(filename, stdin);
   try
      compile filename
   with
     Sys_error msg ->
       output_file();
       print_endline msg;
       raise Bad_end
   | Failure s ->
       output_file();
       print_endline s;
       raise Bad_end
;;
