(* Common part between the GOI and question stack machines *)

(* Code generation *)

(* This program generates a temporary C file, to compile with gcc *)

open Globals;;

let rec nbr x = function
    Apply(e1,e2) -> (nbr x e1) + (nbr x e2)
  | Func(v,e) -> if v = x then 0 else nbr x e
  | Int _ -> 0
  | Var v -> if x = v then 1 else 0
  | F(_,e) -> nbr x e
  | Op(_,e1,e2) -> (nbr x e1) + (nbr x e2)
  | Rec(v,e) -> if v = x then 0 else nbr x e
  | Cond(be,tte,ffe) -> (nbr x be) + (nbr x tte) + (nbr x ffe)
  | Affectin(v,e1,e2) -> (nbr x e1) + (if x = v then 0 else nbr x e2)
;;  

let codepar e1 e2 =
  let l3out = new_label() in
  output_string !oc ("
     R0 = (R0 << 1) + 1;
     latexout(\"q\");\n"
     ^ l3out ^ ":");
  e1();
  output_string !oc ("
     if (R0 & 1) {
        R0 >>= 1;
        latexout(\"q*\");
     } else {
        R0 >>= 1;
        latexout(\"p*\");");
  e2();
  output_string !oc ("
        R0 <<= 1;
        latexout(\"p\");
        goto " ^ l3out ^ ";
     };")
;;

let codeapply e1 e2 =
  let l3out = new_label() in
  output_string !oc ("
     R0 = (R0 << 1) + 1;
     latexout(\"q\");\n"
     ^ l3out ^ ":
     _DER");
  e1();
  output_string !oc ("
     _DERREV
     if (R0 & 1) {
        R0 >>= 1;
        latexout(\"q*\");
     } else {
        R0 >>= 1;
        latexout(\"p*\");");
  e2();
  output_string !oc ("
        R0 <<= 1;
        latexout(\"p\");
        goto " ^ l3out ^ ";
     }")
;;

let codefun backvarref e =
  output_string !oc "
     ep = ep->right;
     latexout(\"!\");
     if (R0 & 1) {
        R0 >>= 1;
        latexout(\"q*\");";
  e();
  output_string !oc "
        R0 = (R0 << 1) + 1;
        latexout(\"q\");
     } else {
        R0 >>= 1;
        latexout(\"p*\");";
  backvarref();
  output_string !oc "
        R0 <<= 1;
        latexout(\"p\");
     };
     ep = ep->prev;
     latexout(\"!*\");"
;;

let codeweakfun e =
  output_string !oc "
     ep = ep->right;
     latexout(\"!\");
     R0 >>= 1;
     latexout(\"q*\");";
  e();
  output_string !oc "
     R0 = (R0 << 1) + 1;
     latexout(\"q\");
     ep = ep->prev;
     latexout(\"!*\");"
;;

let codeerr mess =
  output_string !oc ("
    printf(\"" ^ mess ^ "\\n\");
    latexend();
    latexendall();
    exit(1);\n")
;;

let putqask expr =
  if !mach = Question then
  begin
     let l2 = new_label() in
     output_string !oc ("
     _QASK(" ^ l2 ^ ")");
     expr();
     output_string !oc ("\n"
     ^ l2 ^ ":")
  end
  else
     expr()
;;

let putqaskns expr =
  if !mach = Question then
  begin
     let l2 = new_label() in
     output_string !oc ("
     _QASKNS(" ^ l2 ^ ")");
     expr();
     output_string !oc ("\n"
     ^ l2 ^ ":")
  end
  else
     expr()
;;

let codecond boolexpr ttexpr ffexpr =
  putqask boolexpr;
  output_string !oc ("
     if (s_top(vstack)) {
        s_pop(vstack);
        latexout(\"cond\");");
  ttexpr();
  output_string !oc "
     } else {
        s_pop(vstack);
        latexout(\"cond\");";
  ffexpr();
  output_string !oc "
     }"
;;

let codenil () =
  output_string !oc "
     if (s_top(dstack2) == 1) {
        s_pop(dstack2);
        latexout(\"i*\");
        s_push(vstack,tt);
        _QANS(\"tt\");
     } else {
        printf(\"Bad access to nil node\\n\");
        latexend();
        latexendall();
        exit(0);
     }"
;;

let codelistask c appexpr =
  if c = "tl" then
    output_string !oc "
     s_top(dstack2)+=2;
     latexout(\"l\");"
  else
    output_string !oc ("
     s_push(dstack2, " ^ (if c = "hd" then "0" else "1") ^ ");
     latexout(\"" ^ (if c = "hd" then "h" else "i") ^ "\");");
  appexpr()
;;

let codecst s =
  output_string !oc ("
    s_push(vstack, " ^ s ^ ");
    _QANS(\"" ^ s ^"\");") 
;;
       
let codeinternal name appexpr =
  putqaskns appexpr;
  output_string !oc ("\n     " ^
      (match name with
          "iszero" | "not" -> "s_top(vstack) = (s_top(vstack) == 0)"
        | "succ" -> "s_top(vstack)++"
        | "pred" -> "s_top(vstack)--"
        | _ -> failwith "Internal error : undefined language identifier")
    ^ ";
    _QANS(\"" ^ name ^"\");")
;;

let codeop name e1 e2 =
match name with
  "::" ->
  output_string !oc "
     if (s_top(dstack2) > 1) {
       s_top(dstack2)-=2;
       latexout(\"l*\");";
  e2();
  output_string !oc "
     } else if (s_top(dstack2) == 1) {
       s_pop(dstack2);
       latexout(\"i*\");
       s_push(vstack,ff);
       _QANS(\"ff\");
     } else {
       s_pop(dstack2);
       latexout(\"h*\");";
  e1();
  output_string !oc "
     }"
| "and" ->
  putqask e1;
  output_string !oc ("
    if (s_top(vstack)) {
      s_pop(vstack);
      latexout(\"and\");");
  e2();
  output_string !oc "
    }"
| "or" ->
  putqask e1;
  output_string !oc ("
    if (s_top(vstack) == 0) {
      s_pop(vstack);
      latexout(\"or\");");
  e2();
  output_string !oc "
    }"
| _ ->
  putqask e1;
  putqaskns e2;
  output_string !oc ("
    { signed long tmp = s_top(vstack);
      s_pop(vstack);
      s_top(vstack) = (s_top(vstack) " ^ (match name with 
        "=" -> "=="
      | "<>" | "xor" -> "!="
      | _ -> name)          
      ^ " tmp);
    }
    _QANS(\"" ^ (match name with
         "%" -> "\\\\%"
       | "<=" -> "$<=$"
       | "<" -> "$<$"
       | "<>" -> "$<>$"
       | _    ->  name) ^"\");")  
;;

type varcode_t = string * string * int * int ref;;

type symbtable_t = (string * varcode_t) list;;

let globaldecl = ref ([] : (string * sexpr * symbtable_t * varcode_t) list);;
 
(* ident,(in_label, out_label, box_level, ref number_of_calls) *)
(* special identifiers
   are coded by their corresponding node on EACH call, instead of making
   contractions and box doors, if they are immediately applied to arguments.
   Else, they are coded according to expressions given in globaldecl
Careful !!
   The number of calls must be calculated before the variable in defined
   or referenced in the C code (because the code is different if there is
   one or several calls)
 *)

let globalsym = ref ([] : symbtable_t)
;;

let box_level = ref 0;;

(* Access to a variable 
   lev is the level where the variable is called
   level is the level where the variable is defined
*)

let putquestions lev level myvar =
   (* when a D node meets the (?t) chain *)
   if lev < 0 then 
      if abs(lev)-1 = level then
      begin
         output_string !oc "\n     _DER";
         myvar();
         output_string !oc "\n     _DERREV"
      end
      else if abs(lev)-2 = level then
      begin
         output_string !oc "\n     ep = ep->prev;";
         myvar();
         output_string !oc "\n     ep = ep->right;"
      end
      else
      begin
         output_string !oc ("
     ep = ep->prev;
     { env_t *tmp = ep;
       ");
         if abs(lev)-2 = level + 1 then 
            output_string !oc "tmp = tmp->prev;"
         else 
            output_string !oc ("int i;
       for (i=1; i<=" ^ string_of_int (abs(lev)-2-level) ^ "; i++)
          tmp = tmp->prev;");
         output_string !oc ("
       _QUESTION
     }");
         myvar();
         output_string !oc ("
     { env_t *tmp = ep->left;
       ");
         if abs(lev)-2 > level + 1 then
            output_string !oc ("int i;
       for (i=1; i<" ^ string_of_int (abs(lev)-2-level) ^ "; i++)
          tmp = tmp->right;");
         output_string !oc ("
       _QUESTIONREV
     }
     ep = ep->right;")
      end
   else if lev <> level then
   begin
         output_string !oc ("
     { env_t *tmp = ep;
       ");
         if lev = level + 1 then 
            output_string !oc "tmp = tmp->prev;"
         else 
            output_string !oc ("int i;
       for (i=1; i<=" ^ string_of_int (lev-level) ^ "; i++)
          tmp = tmp->prev;");
         output_string !oc ("
       _QUESTION
     }");
         myvar();
         output_string !oc ("
     { env_t *tmp = ep->left;
       ");
         if lev > level + 1 then
            output_string !oc ("int i;
       for (i=1; i<" ^ string_of_int (lev-level) ^ "; i++)
          tmp = tmp->right;");
         output_string !oc ("
       _QUESTIONREV
     }")
  end
  else
     myvar()
;;

let oncevar (lblin, lblout) =
   output_string !oc ("
     goto " ^ lblout ^ ";\n"
     ^ lblin ^ ":")
;;      

(* Access to a variable used several times *)

let multvar (_, lblout) =
   let liout = new_label() in
   output_string !oc ("
     _CONT(&&" ^ liout ^ ")
     latexout(\"r\");
     goto " ^ lblout ^ ";\n"
     ^ liout ^ ":")
;;

let varref callev (lblin,lblout,level,n) =
   if !n = 1 then
       putquestions callev level (fun () -> oncevar (lblin,lblout))
   else
       putquestions callev level (fun () -> multvar (lblin,lblout))
;;

let backtovarref (lblin,lblout,level,n) =
  if !n = 1 then
    output_string !oc ("
     goto " ^ lblin ^ ";\n"
     ^ lblout ^ ":")
  else
    output_string !oc ("
     _CONTRET\n"
     ^ lblout ^ ":")
;;

let rec incrnumbers lsym e0 = match e0 with
   Apply(e1,e2) -> incrnumbers lsym e1; incrnumbers lsym e2
 | Rec(x,e) -> incrnumbers ((x,("","",0,ref 1))::lsym) e
 | Cond(be,tte,ffe) -> incrnumbers lsym be;
                       incrnumbers lsym tte;
                       incrnumbers lsym ffe
 | F(_,e) -> incrnumbers lsym e
 | Op(_,e1,e2) -> incrnumbers lsym e1; incrnumbers lsym e2
 | Func(x,e) -> incrnumbers ((x,("","",0,ref 0))::lsym) e
 | Affectin(x,e1,e2) -> incrnumbers lsym e1;
                        incrnumbers ((x,("","",0,ref 0))::lsym) e2
 | Var x -> (match x with
             "tt" | "ff" | "[]" | "_match_error" -> ()
            | _ -> try
                     (match List.assoc x lsym with
                         (_,_,_,n) -> incr n)
                   with 
                      Not_found -> raise (Error (e0,"Variable " ^ x ^ 
                        " not found")))
 | Int _ -> ()
;;

let rec calcnumbers = function
   [] -> ()
 | ((x,e,lsym,(_,_,_,n))::ldecl) -> 
      if !n>0 then 
      begin
         try
            incrnumbers lsym e
         with
            Error(e0, msg) ->
               print_underline (Affect(x,e)) e0;
               print_endline msg;
               raise Bad_end
      end;
      calcnumbers ldecl
;; 

let rec nbraff ty = 
   match  (Typing.instanvar ty).type_desc with
     TConst "int" | TConst "bool" -> 1
   | TList t -> 1 + nbraff t
   | _ -> 0
;;

let rec code_expr symtable e0 = match e0 with
  Func(x,e) -> let n = nbr x e in
                  incr box_level;
                  if n = 0 then
                     codeweakfun (fun () -> code_expr symtable e)
                  else 
                  begin
                     let varcode = (new_var_code x,new_var_code x,!box_level,ref n)
                     in codefun (fun () -> backtovarref varcode) 
                          (fun () -> code_expr ((x, varcode) :: symtable) e)
                  end;
                  decr box_level
| F(x,e2) ->
     begin
         match x with
               "hd" | "tl" | "isnil" ->
                    codelistask x (fun () -> code_expr symtable e2)
             | _ -> 
                    codeinternal x (fun () -> code_expr symtable e2)
     end
| Rec(phi,e) ->
     let n = nbr phi e in
         if n = 0 then 
         begin
             print_endline "Internal error : useless rec \
                 should have been removed by optimization";
             code_expr symtable e
         end
         else
             let lblout = new_var_code phi in
             let varcode = ("", lblout, !box_level, ref (n+1))
             and liout = new_var_code phi
             in output_string !oc ("
     _CONT(&&" ^ liout ^ ")
     latexout(\"r\");\n"
     ^ lblout ^ ":");
                code_expr ((phi, varcode)::symtable) e;
                output_string !oc ("
     _CONTRET\n"
     ^ liout ^ ":")
| Op(x, e3, e2) ->
     codeop x (fun () -> code_expr symtable e3)
              (fun () -> code_expr symtable e2)
| Cond(be, e3, e2) ->
     codecond (fun () -> code_expr symtable be)
              (fun () -> code_expr symtable e3)
              (fun () -> code_expr symtable e2)
| Apply(e1,e2) ->
         begin
             match e1 with
               Var x -> (* optimization : if D and ?t nodes meet,
                       they reduce to a !* node, the apply node,
                       which is par-D becomes par only, and the variable
                       call will be treated specially to replace the
                       ?t node by a !* node, that's why it has a negative
                       box level *)
                    (try
                      codepar (fun () -> varref (- !box_level -1)
                        (List.assoc x symtable)) (fun () -> code_expr symtable e2)
                    with 
                      Not_found -> raise (Error(e1, "Variable "^x^
                         " not declared")))
              | _ -> codeapply (fun () -> code_expr symtable e1)
                               (fun () -> code_expr symtable e2)
         end
| Var(x) -> begin
             match x with
                 "tt" | "ff" -> codecst x
               | "_match_error" -> codeerr "Pattern-matching failed"
               | "[]" -> codenil ()
               | _ -> try
                        varref !box_level (List.assoc x symtable)
                      with 
                        Not_found -> raise (Error(e0, "Variable " ^ x ^ 
                          " not declared"))
            end
| Int n -> codecst (string_of_int n)
| Affectin(x,e1,e2) -> let n = nbr x e2 in
              let varcode = (new_var_code x,new_var_code x,!box_level,ref n)
              in code_expr ((x, varcode)::symtable) e2;
                 if n<>0 then
                 (* add e1 in globaldecl so that it is coded later *)
                    globaldecl := (x, e1, symtable, varcode) :: !globaldecl
;;

let rec code_prog () =
   match !globaldecl with 
       [] -> ()
    |  ((x, e, syml, (lblin, lblout, level, n)) :: decll) -> 
         globaldecl := decll;
         if !n > 0 then
         begin
            output_string !oc ("\n"
                  ^ lblout ^ ":");
            box_level := level;
            begin
               try
                  code_expr syml e
               with
                  Error(e0, msg) ->
                     print_underline (Affect(x,e)) e0;
                     print_endline msg;
                     raise Bad_end
            end;
            if !n = 1 then
               output_string !oc ("
     goto " ^ lblin ^ ";")
            else
               output_string !oc "
     _CONTRET\n"
         end;
         code_prog ()
;;

let codeend () =

  output_string !oc "
     latexendall();
     return 0;\n\n";
  calcnumbers !globaldecl;
  code_prog ();
  output_string !oc "};\n"
;;

let rec codeaffint varcode =
  output_string !oc "
     _INIT";
  putqask (fun () -> varref 0 varcode);
  output_string !oc "
     latexout(\"printing\");
     printf(\"%li\",s_top(vstack));
     s_pop(vstack);
     h_freeall(henv);"

and

codeaffbool varcode =
  output_string !oc "
     _INIT";
  putqask (fun () -> varref 0 varcode);
  output_string !oc "
     latexout(\"printing\");
     if (s_top(vstack))
        printf(\"tt\");
     else
        printf(\"ff\");
     s_pop(vstack);
     h_freeall(henv);"

and

codeafflist t varcode =
  output_string !oc "
     printf(\"[\");
     s_push(dstack,0);
     do {
          s_top(dstack)++; /* i */
          _INIT";
  putqask (fun () -> varref 0 varcode);
  output_string !oc "
          h_freeall(henv);
          if (s_top(vstack))
             break;
          s_pop(vstack);
          s_top(dstack)--; /* i -> h */
          if (s_top(dstack) != 0)
             printf(\"; \");";
  codeaff t varcode;
  output_string !oc "
          s_top(dstack)+=2;/* one more l */
     } while (1);
     s_pop(vstack);
     s_pop(dstack);
     printf(\"]\");"

and

codeaffdumb varcode =
  output_string !oc "
     latexout(\"printing\");
     printf(\"<function>\");\n"

and codeaff ty varcode = 
               match (Typing.instanvar ty).type_desc with
                   TConst "int" -> codeaffint varcode
                 | TConst "bool" -> codeaffbool varcode
                 | TList t -> codeafflist t varcode
                 | _ -> codeaffdumb varcode
;;

let rec record_prog = function 
   [] -> globaldecl := [];
         globalsym := []
 | (x,ty,e,lbl)::syml ->
  record_prog syml;
  let nbrcalls = ref 0 in 
  let varcode = (new_var_code lbl, new_var_code lbl, 0, nbrcalls) in
    globaldecl := (x, e, !globalsym, varcode) :: !globaldecl;
    globalsym := (x, varcode) :: !globalsym;
    if (String.length x > 5) & (String.sub x 0 5 = "_expr") then
    begin
       nbrcalls := nbraff ty;
       output_string !oc "
      latexstart();
      fputs(\"";
       print_prog !oc (Expr e);
       output_string !oc "Result : \", stdout);\n";
       if !repeat > 1 then 
          output_string !oc ("
      for (r=1; r<=" ^ (string_of_int !repeat) ^ "; r++) {");
       codeaff ty varcode;
       if !repeat > 1 then
          output_string !oc "
      printf(\" - \");
      }";
       output_string !oc "
      printf(\"\\n\");
      latexend();\n\n"
   end
;;

let code symtable =
  label_count := 0;
  if !latex then
     output_string !oc "#define latex\n\n";
  if !count then
     output_string !oc "#define count\n\n";
  output_string !oc ("/* This program was generated by sflcomp */

#include \"" ^ !Config.macrodir ^ "sflconfig.c\"
#include \"" ^ !Config.macrodir ^ "gqnammacros.c\"
#include \"" ^ !Config.macrodir ^ (if !mach = GOI then "g" else "q") ^ "nammacros.c\"

ulng r;

int main()
{");
  if !meml > 0 then
     output_string !oc ("
  putlimit(RLIMIT_DATA, " ^ (string_of_int (!meml * 1024)) ^ ");");
  if !timel > 0 then
     output_string !oc ("
  putlimit(RLIMIT_CPU, " ^ (string_of_int !timel) ^ ");");
  output_string !oc ("
     latexstartall();
     printf(\"" ^ (machinename !mach) ^ " - SFL program running...\\n\");
\n");
  record_prog symtable;
  output_string !oc "
     latexendall();
     return 0;\n\n";
  calcnumbers !globaldecl;
  code_prog ();
  output_string !oc "};\n"
;;

