(* Krivine machine - common part for all call models *)

(* Code generation *)

(* This program generates a temporary C file, to compile with gcc *)

open Globals;;

let codeerr mess =
  output_string !oc ("
    printf(\"" ^ mess ^ "\\n\");
    latexend();
    latexendall();
    exit(1);\n")
;;

type varref = FPar
            | LetVar of string * bool ref;;

let no_fpar l = List.for_all (function (_,x) -> x <> FPar) l;;

(* expr, label, called?, symbol table, box_level *)

let globaldecl = ref ([] : (sexpr * string * bool ref * ((string * varref) list
)) list);;
 
(* ident, varref *)
(* special identifiers
   are coded by their corresponding node on EACH call, instead of making
   contractions and box doors, if they are immediately applied to arguments.
   Else, they are coded according to expressions given in globaldecl *)

let globalsym = ref ([] : (string * varref) list);;

let codeaffdumb () =
  output_string !oc "
     latexout(\"printing\");
     printf(\"<function>\");"
;;

let code (incl, code_fun, codecond, code_xref, code_letref, code_globref,
   codenil, codelistask, codecst, codeinternal, codeop, codelet, 
   codeaffint, codeaffbool, codeafflist, splitenv)
   symtabl  =

let rec codeaff lblin ty called = 
               match (Typing.instanvar ty).type_desc with
                   TConst "int" -> codeaffint lblin called
                 | TConst "bool" -> codeaffbool lblin called
                 | TList t -> codeafflist lblin (codeaff lblin t) called
                 | _ -> codeaffdumb ()

in

let rec record_prog = function 
   [] -> globalsym := [];
         globaldecl := []
 | (x,ty,e,l)::syml ->
  record_prog syml;
  let called = ref false 
  and lbl = new_var_code l in
     globaldecl := (e, lbl, called, !globalsym) :: !globaldecl;
     if (String.length x <= 5) or (String.sub x 0 5 <> "_expr") then
        globalsym := (x, LetVar (lbl,called)) :: !globalsym 
     else
     begin
        output_string !oc "
     latexstart();
     fputs(\"";
        print_prog !oc (Expr e);
        output_string !oc "Result : \", stdout);\n";
        if !repeat > 1 then 
           output_string !oc ("
     for (r=1; r<=" ^ (string_of_int !repeat) ^ "; r++) {");
        codeaff lbl ty called;
        if !repeat > 1 then
           output_string !oc "
     printf(\" - \");
     }";
        output_string !oc "
     printf(\"\\n\");
     latexend();\n\n"
     end

in

let rec code_expr symtable e0 = match e0 with
  Func(x,e) -> code_fun ();
               code_expr ((x, FPar) :: symtable) e
| F(x,e2) ->
    begin
        match x with
               "hd" | "tl" | "isnil" ->
                    codelistask (match x with 
                      "hd" -> "h"
                    | "tl" -> "l"
                    | "isnil" -> "i"
                    | _ -> failwith "internal error: unknown list op");
                    code_expr symtable e2
             | _ ->
                    codeinternal x (fun () -> code_expr symtable e2)
     end
| Rec(phi,e) ->
     let lbl = new_label() in
         output_string !oc ("\n" ^ lbl ^ ":");
         code_expr ((phi, LetVar(lbl, ref true)) :: symtable) e
| Op(x, e3, e2) ->
     codeop x code_expr symtable e3 e2
| Cond(be, e3, e2) ->
     codecond code_expr symtable be e3 e2
| Apply(e1, e2) ->
     let l = new_label() 
     in let (s1,s2) =  splitenv e1 e2 l symtable in
        (match !mach with
           IrrevNeed(_) -> (match e2 with
               Func(_,_) | Op("::",_,_) | Var "[]" -> output_string !oc "
     ep->l = _cv;"
             | _ -> ())
         | _ -> ());
        output_string !oc "
     latexout(\"d\");";
        code_expr s1 e1;
        output_string !oc ("\n" ^ l ^ ":");
        code_expr s2 e2
| Var(x) -> begin
              match x with
                 "tt" | "ff" -> codecst x
               | "_match_error" -> codeerr "Pattern-matching failed"
               | "[]" -> codenil ()
               | _ -> 
let rec codevarref deltalev = function 
   [] -> raise (Error (e0, "Variable " ^ x ^ " not declared"))
 | ((y, FPar)::l) -> 
      if x = y then
         code_xref deltalev
      else
         codevarref (deltalev+1) l
 | ((y, LetVar (lbl,called))::l) -> 
      if x = y then 
         if no_fpar l then
            code_globref lbl called
         else
            code_letref deltalev lbl called
      else
         codevarref deltalev l

in
  codevarref 0 symtable
            end
| Int n -> codecst (string_of_int n)
| Affectin(x,e1,e2) -> 
    codelet x code_expr symtable e1 e2

in

let rec code_prog = function
      [] -> ()
    | (e,lbl,called,symb)::gl ->
         if (!called) then
         begin
            output_string !oc ("\n" ^ lbl ^ ":");
            try
               code_expr symb e
            with
               Error(e0, msg) ->
                 print_underline (Expr e) e0;
                 print_endline msg;
                 raise Bad_end
         end;
         code_prog gl

in

  label_count := 0;
  if !latex then
     output_string !oc "#define latex\n\n";
  if !count then
     output_string !oc "#define count\n\n";
  output_string !oc "/* This program was generated by sflcomp */\n";
  List.iter (fun f -> output_string !oc ("
#include \"" ^ !Config.macrodir ^ f ^ "\"")) incl;
  if !repeat > 1 then
     output_string !oc "\n
ulng r;";
  output_string !oc "\n
int main()
{";
  if !meml > 0 then
     output_string !oc ("
  putlimit(RLIMIT_DATA, " ^ (string_of_int (!meml * 1024)) ^ ");");
  if !timel > 0 then
     output_string !oc ("
  putlimit(RLIMIT_CPU, " ^ (string_of_int !timel) ^ ");");
  output_string !oc ("
     latexstartall();
     printf(\"" ^ (machinename !mach) ^ " - SFL program running...\\n\");
\n");
  record_prog symtabl;
  output_string !oc "     latexendall();
     return 0;\n\n";
  code_prog !globaldecl;
  output_string !oc "};\n"
;;

