(* Krivine machine *)

(* Code generation *)

(* This program generates a temporary C file, to compile with gcc *)

open Globals;;
open Kall;;

let code_fun ()=
  output_string !oc "
     { env_t *tmp = ep;
       ep = ep->next;
       tmp->next = context;
       context = tmp;
     }
     latexout(\"!\");"
;;

let nosplitenv e1 e2 l1out symtable =
  output_string !oc ("
     _DER(" ^ l1out ^ ")");
  (symtable,symtable)
;;

let codecond code_expr symtable boolexpr ttexpr ffexpr =
  let l2 = new_label() in
  output_string !oc ("
     _DER(" ^ l2 ^ ");
     latexout(\"cond ask\");");
  code_expr symtable boolexpr;
  output_string !oc ("\n"
     ^ l2 ^ ":
     if (s_top(vstack)) {
        s_pop(vstack);
        latexout(\"cond ans\");");
  code_expr symtable ttexpr;
  output_string !oc "
     } else {
        s_pop(vstack);
        latexout(\"cond ans\");";
  code_expr symtable ffexpr;
  output_string !oc "
     }"
;;

let gccode_xref n =
  output_string !oc "
    { env_t *tmp = context;
      void *tmpad;"; 
  if n=1 then
     output_string !oc "
      context = context->next;"
  else if n>1 then
     output_string !oc ("
      { short i;
        for (i=1; i<=" ^ (string_of_int n) ^ "; i++)
            context = context->next;
      }");
  output_string !oc ("
      tmpad = context->addr;
      context = context->cont;
      if (context != NIL) context->nbrcopy++;
      freetree(tmp);
      latexout(\"" ^ (string_of_int n) ^ " ref\");
      goto *tmpad;
    }\n")
;;

let gccode_letref n lg called = 
  if n>0 then
  begin
     output_string !oc "
    { env_t *tmp = context;";
     if n=1 then
        output_string !oc "
      context = context->next;"
     else if n=2 then
        output_string !oc "
      context = context->next->next;"
     else if n>2 then
        output_string !oc ("
      { short i;
        for (i=0; i<" ^ (string_of_int n) ^ "; i++)
            context = context->next;
      }");
     output_string !oc "
      if (context != NIL) context->nbrcopy++;
      freetree(tmp);
    }"
  end;
  output_string !oc ("
    latexout(\"" ^ (string_of_int n) ^ " let ref\");
    goto " ^ lg ^ ";\n");
  called := true
;;

let cgccode_globref lg called =
  output_string !oc ("
      freetree(context);
      context = NIL;
      latexout(\"global ref\");
      goto " ^ lg ^ ";\n");
  called := true
;;

let codenil () =
  output_string !oc ("
     if (s_top(dstack2) == 1) {
        s_pop(dstack2);
        latexout(\"i*\");
        s_push(vstack,tt);
        _DERREV(\"tt\");
     } else {
        printf(\"Bad access to nil node\\n\");
        latexend();
        latexendall();
        exit(0);
     }\n\n")
;;

let codelistask c =
  if c = "l" then
    output_string !oc "
     s_top(dstack2) += 2;
     latexout(\"l\");"
  else
    output_string !oc ("
     s_push(dstack2, " ^ (if c = "h" then "0" else "1") ^ ");
     latexout(\"" ^ c ^ "\");")
;;

let codecst s =
  output_string !oc ("
    s_push(vstack, " ^ s ^ ");
    _DERREV(\"" ^ s ^ "\")\n") 
;;
       
let codeinternal name appexpr =
  let l2 = new_label() in
  output_string !oc ("
    _DERNS(" ^ l2 ^ ",\"" ^ name ^ " ask\")");
  appexpr();
  output_string !oc ("\n"
    ^ l2 ^ ":
    " ^
        (match name with
            "iszero" | "not" -> "s_top(vstack) = (s_top(vstack) == 0)"
          | "succ" -> "s_top(vstack)++"
          | "pred" -> "s_top(vstack)--"
          | _ -> failwith "Internal error : undefined language identifier")
    ^ ";
    _DERREV(\"" ^ name ^ " ans\")\n")
;;

let codeop name code_expr symtable e1 e2 =
match name with
  "::" ->
  output_string !oc "
     if (s_top(dstack2) > 1) {
       s_top(dstack2) -= 2;
       latexout(\"l*\");";
  code_expr symtable e2;
  output_string !oc "
     } else {
       if (s_top(dstack2) == 1) {
         s_pop(dstack2);
         latexout(\"i*\");
         s_push(vstack,ff);
         _DERREV(\"ff\");
       } else {
         s_pop(dstack2);
         latexout(\"h*\");";
  code_expr symtable e1;
  output_string !oc "
       }
     }"
| "and" ->
  let l2 = new_label() 
  in output_string !oc ("
    _DER(" ^ l2 ^ ");
    latexout(\"and ask\");");
  code_expr symtable e1;
  output_string !oc ("\n"
    ^ l2 ^ ":
    if (s_top(vstack)) {
      s_pop(vstack);
      latexout(\"and ans\");");
  code_expr symtable e2;
  output_string !oc "
    } else 
      _DERREV(\"and ans\")"
| "or" ->
  let l2 = new_label() 
  in output_string !oc ("
    _DER(" ^ l2 ^ ");
    latexout(\"or ask\");");
  code_expr symtable e1;
  output_string !oc ("\n"
    ^ l2 ^ ":
    if (s_top(vstack) == 0) {
      s_pop(vstack);
      latexout(\"or ans\");");
  code_expr symtable e2;
  output_string !oc "
    } else 
      _DERREV(\"or ans\")"
| _ ->
  let l2 = new_label() 
  and l3 = new_label()
  and texname = match name with
         "%" -> "\\\\%"
       | "<=" -> "$<=$"
       | "<" -> "$<$"
       | "<>" -> "$<>$"
       | _    ->  name
  in output_string !oc ("
    _DER(" ^ l2 ^ ");
    latexout(\"" ^ texname ^ " ask 1\");");
    code_expr symtable e1;
    output_string !oc ("\n"
    ^ l2 ^ ":
    _DERNS(" ^ l3 ^ ",\"" ^ texname ^ " ask 2\")");
    code_expr symtable e2;
    output_string !oc ("\n"
    ^ l3 ^ ":
    { signed long tmp = s_top(vstack);
      s_pop(vstack);
      s_top(vstack) = (s_top(vstack) " ^ (match name with 
        "=" -> "=="
      | "<>" | "xor" -> "!="
      | _ -> name)
          ^ " tmp);
    }
    _DERREV(\"" ^ texname ^ " ans\")\n\n")
;;

let codeaffint lblin called=
  let lblout = new_label() in
  output_string !oc ("
     _INIT(" ^ lblout ^ ")
     goto " ^ lblin ^ ";\n"
     ^ lblout ^ ":
     latexout(\"printing\");
     printf(\"%li\",s_top(vstack));
     s_pop(vstack);
     h_freeall(henv);");
  called := true
;;

let codeaffbool lblin called =
  let lblout = new_label() in
  output_string !oc ("
     _INIT(" ^ lblout ^ ")
     goto " ^ lblin ^ ";\n"
     ^ lblout ^ ":
     latexout(\"printing\");
     if (s_top(vstack))
        printf(\"tt\");
     else
        printf(\"ff\");
     s_pop(vstack);
     h_freeall(henv);");
   called := true
;;

let codeafflist lblin affelem called =
  called := true;
  let lblout = new_label() in
  output_string !oc ("
     printf(\"[\");
     s_push(dstack,0);
     do {
          s_top(dstack)++; /* i */
          _INIT(" ^ lblout ^ ")
          goto " ^ lblin ^ ";\n"
          ^ lblout ^ ":
          h_freeall(henv);
          if (s_top(vstack))
             break;
          s_pop(vstack);
          s_top(dstack)--; /* i -> h */
          if (s_top(dstack) != 0)
             printf(\"; \");");
  affelem called;
  output_string !oc ("
          s_top(dstack)+=2;/* one more l */
     } while (1);
     s_pop(vstack);
     s_pop(dstack);
     printf(\"]\");")
;;

let letcode x code_expr symtable e1 e2 =
       let l =new_var_code x
       and called = ref false
       in code_expr ((x,LetVar (l,called)) :: symtable) e2;
          if !called then
          begin
             output_string !oc ("\n" ^ l ^ ":");
             code_expr symtable e1
          end
;;

let machine = (["sflconfig.c"; "gcknammacros.c"],
   code_fun, codecond, gccode_xref, gccode_letref, cgccode_globref,
   codenil, codelistask, codecst, codeinternal, codeop, letcode, 
   codeaffint, codeaffbool, codeafflist, nosplitenv);;
    
