(* Call by value Krivine machine *)

(* Code generation *)

(* This program generates a temporary C file, to compile with gcc *)

open Globals;;
open Kall;;

let code_fun () =
  let l2out = new_label()
  and l3 = new_label() in
  output_string !oc (l3 ^ ":
      {
         void *tmpad = ep->v.addr;
         if (ep->l == _L) {
            ep->l = 0;
            ep->v.addr = &&" ^ l2out ^ ";
            _SWAPCONT
         } else {
            ep->v.addr = &&" ^ l3 ^ ";
            _SWAP2
         }
         latexout(\"!\");
         goto *tmpad;
      }\n" ^ l2out ^ ":")
;;

let codecond code_expr symtable boolexpr ttexpr ffexpr =
  let l2 = new_label() in
  output_string !oc ("
     _DER(" ^ l2 ^ ")
     latexout(\"cond ask\");");
  code_expr symtable boolexpr;
  output_string !oc ("\n"
     ^ l2 ^ ":
     { signed long tmpval = context->v.val;
       _POPCONTEXTVALUE
       latexout(\"cond ans\");
       if (tmpval) {");
  code_expr symtable ttexpr;
  output_string !oc "
       } else {";
  code_expr symtable ffexpr;
  output_string !oc "
       }
     }"
;;

let gccode_xref n =
  output_string !oc "
    { env_t *tmp = context;";
  if n=1 then
     output_string !oc "
      context = context->next;"
  else if n>1 then
     output_string !oc ("
      { short i;
        for (i=1; i<=" ^ (string_of_int n) ^ "; i++)
            context = context->next;
      }");
  output_string !oc ("
      if (context->l == _v) {
        signed long tmpval = context->v.val;
        context = tmp;
        _PUTVAL(tmpval,\"" ^ (string_of_int n) ^ " ref\")
      } else {
        void *tmpad = context->v.addr;
        context = context->cont;
        if (context != NIL) context->nbrcopy++;
        freetree(tmp);
        latexout(\"" ^ (string_of_int n) ^ " ref\");
        goto *tmpad;
      }
    }")
;;

let codenil () =
  let l = new_label() in
  output_string !oc ("\n"
    ^ l ^ ":
    if (ep->l == 0) 
       {
         void *tmpad = ep->v.addr;
         ep->v.addr = &&" ^ l ^ ";
         _SWAP2
         latexout(\"nil\");
         goto *tmpad;
      }
    else if (ep->l == _i) {
      _DERREV
      _PUTVAL(tt,\"nil-tt\")
    } else {
      printf(\"Bad access to nil node\\n\");
      latexend();
      latexendall();
      exit(0);
    }\n\n")
;;

let gccodelistask x =
  output_string !oc ("
     { env_t *a;
       h_malloc(henv,a);
       a->next = ep;
       ep = a;
       ep->nbrcopy = 1;
       ep->cont = NIL;
       ep->l = _" ^ x ^ ";
     }
    latexout(\"" ^ x ^ "\");")
;;

let codecst s =
  output_string !oc ("
    _PUTVAL(" ^ s ^ ",\"" ^ s ^ "\")\n\n") 
;;
       
let codeinternal name appexpr =
  let l2 = new_label() in
  output_string !oc ("
    _DERNS(" ^ l2 ^ ")
    latexout(\"" ^ name ^ " ask\");");
  appexpr();
  output_string !oc ("\n"
    ^ l2 ^ ":
    { signed long tmpval = context->v.val" ^
        (match name with
            "iszero" | "not" -> " == 0"
          | "succ" -> " + 1"
          | "pred" -> " - 1"
          | _ -> failwith "Internal error : undefined language identifier")
    ^ ";
      _PUTVAL(tmpval,\"" ^ name ^ " ans\")
    }")
;;

let codeop name code_expr symtable e1 e2 =
match name with
  "::" ->
  let l1 = new_label() in
  output_string !oc ("\n"
     ^ l1 ^ ":
     switch (ep->l) {
      case _h :
            { _DERREV
              latexout(\"cons\");");
  code_expr symtable e1;
  output_string !oc "
            }
      case _l :
            { _DERREV
              latexout(\"cons\");";
  code_expr symtable e2;
  output_string !oc (" 
            }
      case _i :
            { _DERREV
              _PUTVAL(ff,\"cons-ff\")
            }
      case 0 :
          {  void *tmpad = ep->v.addr;
             ep->v.addr = &&" ^ l1 ^ ";
             _SWAP2
             latexout(\"cons\");
             goto *tmpad;
          }
     }")
| "and" ->
  let l2 = new_label() in
  output_string !oc ("
     _DER(" ^ l2 ^ ")
     latexout(\"and ask\");");
  code_expr symtable e1;
  output_string !oc ("\n"
     ^ l2 ^ ":
     if (context->v.val) {
       _POPCONTEXTVALUE
       latexout(\"and ans\");");
  code_expr symtable e2;
  output_string !oc "
     } else {
       void *tmpad = ep->v.addr;
       _DERREV
       latexout(\"and ans\");
       goto *tmpad;
     }"
| "or" ->
  let l2 = new_label() in
  output_string !oc ("
     _DER(" ^ l2 ^ ")
     latexout(\"or ask\");");
  code_expr symtable e1;
  output_string !oc ("\n"
     ^ l2 ^ ":
     if (context->v.val == 0) {
       _POPCONTEXTVALUE
       latexout(\"or ans\");");
  code_expr symtable e2;
  output_string !oc "
     } else {
       void *tmpad = ep->v.addr;
       _DERREV
       latexout(\"or ans\");
       goto *tmpad;
     }"
| _ ->
  let l2 = new_label()
  and l3 = new_label()
  and texname = match name with
         "%" -> "\\\\%"
       | "<=" -> "$<=$"
       | "<" -> "$<$"
       | "<>" -> "$<>$"
       | _    ->  name
  in output_string !oc ("
    _DER(" ^ l2 ^ ")
    latexout(\"" ^ texname ^ " ask 1\");");
  code_expr symtable e1;
  output_string !oc ("\n"
    ^ l2 ^ ":
    _ENVTOSTACK
    _DERNS(" ^ l3 ^ ")
    latexout(\"" ^ texname ^ " ask 2\");");
  code_expr symtable e2;
  output_string !oc ("\n"
    ^ l3 ^ ":
    { signed long tmpval = (ep->v.val) " ^ (match name with 
        "=" -> "=="
      | "<>" | "xor" -> "!="
      | _ -> name)
          ^ " (context->v.val);
      _DERREV
      _PUTVAL(tmpval,\"" ^ texname ^ " ans\")
    }")
;;

let codelet x code_expr symtable e1 e2 =
   match e1 with
      Func(_,_) | Op("::",_,_) | Var "[]" ->
           (* If expression e1 cannot be simplified, it's useless to keep
              a cell in the environment to store its value; call-by-name
              is better in this case *)
           Knam.letcode x code_expr symtable e1 e2 
    | _ -> 
   let l = new_label() in
   output_string !oc ("
    _DER(" ^ l ^ ")
    ep->l = 0;
    latexout(\"let decl\");");
   code_expr symtable e1;
   output_string !oc ("\n" ^ l ^ ":");
   code_expr ((x, FPar)::symtable) e2
;;

let codeaffint lblin called=
  let lblout = new_label() in
  output_string !oc ("
     _INIT(" ^ lblout ^ ")
     goto " ^ lblin ^ ";\n"
     ^ lblout ^ ":
     latexout(\"printing\");
     printf(\"%li\",context->v.val);
     h_freeall(henv);");
  called := true
;;

let codeaffbool lblin called =
  let lblout = new_label() in
  output_string !oc ("
     _INIT(" ^ lblout ^ ")
     goto " ^ lblin ^ ";\n"
     ^ lblout ^ ":
     latexout(\"printing\");
     if (context->v.val)
        printf(\"tt\");
     else
        printf(\"ff\");
     h_freeall(henv);");
  called := true
;;

let codeafflist lblin affelem called =
  called := true;
  let lblout = new_label() in
  output_string !oc ("
     printf(\"[\");
     s_push(dstack,0);
     do {
          s_top(dstack)++; /* i */
          _INIT(" ^ lblout ^ ")
          goto " ^ lblin ^ ";\n"
          ^ lblout ^ ":
          if (context->v.val)
             break;
          h_freeall(henv);
          s_top(dstack)--; /* i -> h */
          if (s_top(dstack) != 0)
             printf(\"; \");");
  affelem called;
  output_string !oc ("
          s_top(dstack)+=2;/* one more l */
     } while (1);
     h_freeall(henv);
     s_pop(dstack);
     printf(\"]\");")
;;

let machine = (["sflconfig.c"; "gckvalmacros.c"],
    code_fun, codecond, gccode_xref, Knam.gccode_letref, Knam.cgccode_globref,
    codenil, gccodelistask, codecst, codeinternal, codeop,
    codelet,codeaffint, codeaffbool, codeafflist, Knam.nosplitenv);;
    
