(* Call by value Krivine machine *)

(* Code generation *)

(* This program generates a temporary C file, to compile with gcc *)

open Globals;;
open Kall;;

let code_fun () =
  let l2out = new_label()
  and l3 = new_label() in
  output_string !oc (l3 ^ ":
      {
         void *tmpad = ep->v.addr;
         _SWAPCONT
         if (ep->l == _L) {
            ep->l = 0;
            ep->v.addr = &&" ^ l2out ^ ";
         } else {
            ep->v.addr = &&" ^ l3 ^ ";
            ep = ep->right;
         }
         latexout(\"!\");
         goto *tmpad;
      }\n" ^ l2out ^ ":")
;;

let codecond code_expr symtable boolexpr ttexpr ffexpr =
  let l2 = new_label() in
  output_string !oc ("
     _DER(" ^ l2 ^ ")
     latexout(\"cond ask\");");
  code_expr symtable boolexpr;
  output_string !oc ("\n"
     ^ l2 ^ ":
     { signed long tmpval = ep->prev->v.val;
       _POPCONTEXTVALUE
       latexout(\"cond ans\");
       if (tmpval) {");
  code_expr symtable ttexpr;
  output_string !oc "
       } else {";
  code_expr symtable ffexpr;
  output_string !oc "
       }
     }"
;;

let code_xref n =
  if n=0 then 
     output_string !oc "
    { env_t *tmp = ep->prev, *tmp2 = tmp->left;"
  else if n=1 then
     output_string !oc "
    { env_t *tmp = ep->prev->prev, *tmp2 = tmp->left;"
  else
     output_string !oc ("
    { env_t *tmp = ep, *tmp2;
      short i;
      
      for (i=0; i<=" ^ (string_of_int n) ^ "; i++)
          tmp = tmp->prev;
      tmp2 = tmp->left;");
  output_string !oc ("
      if (tmp->l == _v) {
        signed long tmpval = tmp->v.val;
        _PUTVAL(tmpval,\"" ^ (string_of_int n) ^ " ref\")
      } else {
        void *tmpad = tmp->v.addr;
        tmp->left = NIL; /* Disconnect the restored env form the rest */
        putenv(tmp2);
        latexout(\"" ^ (string_of_int n) ^ " ref\");
        goto *tmpad;
      }
    }")
;;

let codenil () =
  let l = new_label() in
  output_string !oc ("\n"
    ^ l ^ ":
    if (ep->l == 0) 
       {
         void *tmpad = ep->v.addr;
         _SWAPCONT
         ep->v.addr = &&" ^ l ^ ";
         ep = ep->right;
         latexout(\"nil\");
         goto *tmpad;
      }
    else if (ep->l == _i) {
      _DERREV
      _PUTVAL(tt,\"nil-tt\")
    } else {
      printf(\"Bad access to nil node\\n\");
      latexend();
      latexendall();
      exit(0);
    }\n\n")
;;

let codelistask x =
  output_string !oc ("
     { env_t *a;
       h_malloc(henv,a);
       a->right = ep->right;
       if (a->right != NIL) a->right->prev = a;
       a->left = ep->left;
       a->v = ep->v;
       a->l = ep->l;
       a->prev = ep;
       ep->l = _" ^ x ^ ";
       ep->right = a;
       ep->left = NIL;
     }
    latexout(\"" ^ x ^ "\");")
;;

let codeinternal name appexpr =
  let l2 = new_label() in
  output_string !oc ("
    _DERNS(" ^ l2 ^ ")
    latexout(\"" ^ name ^ " ask\");");
  appexpr();
  output_string !oc ("\n"
    ^ l2 ^ ":
    { signed long tmpval = ep->prev->v.val" ^
        (match name with
            "iszero" | "not" -> " == 0"
          | "succ" -> " + 1"
          | "pred" -> " - 1"
          | _ -> failwith "Internal error : undefined language identifier")
    ^ ";
      _PUTVAL(tmpval,\"" ^ name ^ " ans\")
    }")
;;

let codeop name code_expr symtable e1 e2 =
match name with
  "::" ->
  let l1 = new_label() in
  output_string !oc ("\n"
     ^ l1 ^ ":
     switch (ep->l) {
      case _h :
            { _DERREV
              latexout(\"cons\");");
  code_expr symtable e1;
  output_string !oc "
            }
      case _l :
            { _DERREV
              latexout(\"cons\");";
  code_expr symtable e2;
  output_string !oc (" 
            }
      case _i :
            { _DERREV
              _PUTVAL(ff,\"cons-ff\")
            }
      case 0 :
          {  void *tmpad = ep->v.addr;
             _SWAPCONT
             ep->v.addr = &&" ^ l1 ^ ";
             ep = ep->right;
             latexout(\"cons\");
             goto *tmpad;
          }
     }")
| "and" ->
  let l2 = new_label() in
  output_string !oc ("
     _DER(" ^ l2 ^ ")
     latexout(\"and ask\");");
  code_expr symtable e1;
  output_string !oc ("\n"
     ^ l2 ^ ":
     if (ep->prev->v.val) {
       _POPCONTEXTVALUE
       latexout(\"and ans\");");
  code_expr symtable e2;
  output_string !oc "
     } else {
       void *tmpad = ep->v.addr;
       _DERREV
       latexout(\"and ans\");
       goto *tmpad;
     }"
| "or" ->
  let l2 = new_label() in
  output_string !oc ("
     _DER(" ^ l2 ^ ")
     latexout(\"or ask\");");
  code_expr symtable e1;
  output_string !oc ("\n"
     ^ l2 ^ ":
     if (ep->prev->v.val == 0) {
       _POPCONTEXTVALUE
       latexout(\"or ans\");");
  code_expr symtable e2;
  output_string !oc "
     } else {
       void *tmpad = ep->v.addr;
       _DERREV
       latexout(\"or ans\");
       goto *tmpad;
     }"
| _ ->
  let l2 = new_label()
  and l3 = new_label()
  and texname = match name with
         "%" -> "\\\\%"
       | "<=" -> "$<=$"
       | "<" -> "$<$"
       | "<>" -> "$<>$"
       | _    ->  name
  in output_string !oc ("
    _DER(" ^ l2 ^ ")
    latexout(\"" ^ texname ^ " ask 1\");");
  code_expr symtable e1;
  output_string !oc ("\n"
    ^ l2 ^ ":
    ep = ep->prev;
    _DERNS(" ^ l3 ^ ")
    latexout(\"" ^ texname ^ " ask 2\");");
  code_expr symtable e2;
  output_string !oc ("\n"
    ^ l3 ^ ":
    { signed long tmpval = (ep->v.val) " ^ (match name with 
        "=" -> "=="
      | "<>" | "xor" -> "!="
      | _ -> name)
          ^ " (ep->prev->v.val);
      ep = ep->right;
      _PUTVAL(tmpval,\"" ^ texname ^ " ans\")
    }")
;;

let codeaffint lblin called=
  let lblout = new_label() in
  output_string !oc ("
     _INIT(" ^ lblout ^ ")
     goto " ^ lblin ^ ";\n"
     ^ lblout ^ ":
     latexout(\"printing\");
     printf(\"%li\",ep->prev->v.val);
     h_freeall(henv);");
  called := true
;;

let codeaffbool lblin called =
  let lblout = new_label() in
  output_string !oc ("
     _INIT(" ^ lblout ^ ")
     goto " ^ lblin ^ ";\n"
     ^ lblout ^ ":
     latexout(\"printing\");
     if (ep->prev->v.val)
        printf(\"tt\");
     else
        printf(\"ff\");
     h_freeall(henv);");
  called := true
;;

let codeafflist lblin affelem called =
  called := true;
  let lblout = new_label() in
  output_string !oc ("
     printf(\"[\");
     s_push(dstack,0);
     do {
          s_top(dstack)++; /* i */
          _INIT(" ^ lblout ^ ")
          goto " ^ lblin ^ ";\n"
          ^ lblout ^ ":
          if (ep->prev->v.val)
             break;
          h_freeall(henv);
          s_top(dstack)--; /* i -> h */
          if (s_top(dstack) != 0)
             printf(\"; \");");
  affelem called;
  output_string !oc ("
          s_top(dstack)+=2;/* one more l */
     } while (1);
     h_freeall(henv);
     s_pop(dstack);
     printf(\"]\");")
;;

let machine = (["sflconfig.c"; "kvalmacros.c"],
   code_fun, codecond, code_xref, Knambis.code_letref, Knambis.code_globref,
   codenil, codelistask, Kval.codecst, codeinternal, codeop, Kval.codelet,
   codeaffint, codeaffbool, codeafflist, Knam.nosplitenv);;
    
