(* Main program for the compiler *)

open Globals;;
open Compiler;;

let help_msg()=
    ("sflcomp by Bruno Blanchet 1995\n
    Toplevel : sflcomp -top\n
    Compiler : sflcomp [options] filename\n
    Options :
    -o OutputFile (now " ^ !outname ^ ")
    -e ExeFile (now : " ^ (if !exename = "" then "no executable output"
        else !exename) ^ ")
    -t : activates/desactivates typing (" ^ (if !types then "activated" else
        "unactivated") ^ " now)
    -l : produce code with Latex trace or not (" ^ (if !latex then "yes" else
        "not") ^ " now)
    -c : count/don't count operations on running (" ^ (if !count then "yes" 
        else "not") ^ " now)
    -v : display information messages or not (" ^ (if !verbose then "yes" else
        "not") ^ " now)
    -r N : repeats evaluations N times (now " ^ (string_of_int !repeat) ^ ")
    -goi : activates the GOI machine
    -goinopt : activates the not optimized GOI machine
    -quest : activates the question stack machine
   Irreversible machines with full copy of environments
    -irrevfckriv : activates the Krivine machine (call-by-name irreversible) 
    -irrevfcvalue : activates the call-by-value irreversible machine
   Irreversible machines with copy of environments
    -irrevckriv : activates the Krivine machine (call-by-name irreversible) 
    -irrevcvalue : activates the call-by-value irreversible machine
   Irreversible machines with garbage collector
    -irrevgckriv : activates the Krivine machine (call-by-name irreversible)
    -irrevgcneed : activates the call-by-need irreversible machine
    -irrevgcvalue : activates the call-by-value irreversible machine
   Present machine is " ^ (machinename !mach) ^ "
    -timel N : limits CPU time to N seconds (now " ^ (string_of_int !timel)
          ^ ")
    -meml N : limits data size to N kbytes (now " ^ (string_of_int !meml) ^ ")
    -h -? : displays this help screen\n")
;;

let help() =
  print_string (help_msg())

let main() =
try
  Arg.parse ["-t", Arg.Unit (fun () -> types := not !types), "activates/desactivates typing";
              "-h", Arg.Unit help, "help message";
              "-?", Arg.Unit help, "help message";
              "-l", Arg.Unit (fun () -> latex := not !latex), "produce code with Latex trace or not";
              "-c", Arg.Unit (fun () -> count := not !count), "count/don't count operations on running";
              "-v", Arg.Unit (fun () -> verbose := not !verbose), "display information messages or not";
              "-r", Arg.Int (fun r -> repeat := r), "repeats evaluations N times";
              "-goi", Arg.Unit (fun () -> mach := GOI), "activates the GOI machine";
              "-goinopt", Arg.Unit (fun () -> mach := GOInopt), "activates the not optimized GOI machine";
              "-quest", Arg.Unit (fun () -> mach := Question), "activates the question stack machine";
              "-irrevfckriv", Arg.Unit (fun () -> mach := IrrevKriv(Big)), "activates the Krivine machine (call-by-name irreversible), full copy";
              "-irrevfcvalue", Arg.Unit (fun () -> mach := IrrevValue(Big)), "activates the call-by-value irreversible machine, full copy";
              "-irrevckriv", Arg.Unit (fun () -> mach := IrrevKriv(C)), "activates the Krivine machine (call-by-name irreversible), copy";
              "-irrevcvalue", Arg.Unit (fun () -> mach := IrrevValue(C)), "activates the call-by-value irreversible machine, copy";
              "-irrevgckriv", Arg.Unit (fun () -> mach := IrrevKriv(GC)), "activates the Krivine machine (call-by-name irreversible), gc";
              "-irrevgcneed", Arg.Unit (fun () -> mach := IrrevNeed(GC)), "activates the call-by-need irreversible machine, gc";
              "-irrevgcvalue", Arg.Unit (fun () -> mach := IrrevValue(GC)), "activates the call-by-value irreversible machine, gc";
              "-top", Arg.Unit Toplevel.main, "toplevel";
              "-timel", Arg.Int (fun n -> timel := n), "limits CPU time to N seconds";
              "-meml", Arg.Int (fun n -> meml := n), "limits data size to N kbytes";
              "-o", Arg.String (fun s -> outname := s), "OutputFile";
              "-e", Arg.String (fun s -> exename := s), "ExeFile";
              "-I", Arg.String (fun s -> Config.macrodir := s), "(undocumented)"
              ] (secure compile) ("sflcomp by Bruno Blanchet 1995\n")
with
   Bad_end -> exit 2
;;

Printexc.catch main();
exit 0;;
