(* Version with link list to memorize contractions,
   to allow unlimited number of nested contractions
   Standard GOI machine *)

(* Code generation *)

(* This program generates a temporary C file, to compile with gcc *)

open Globals;;

let label_count = ref 0;;

let new_var_code x = 
   incr label_count;
   (x ^ "_" ^ (string_of_int !label_count) ^ "_in", 
    x ^ "_" ^ (string_of_int !label_count) ^ "_out")
;;

let new_label() = new_var_code "_lbl";;

let rec codeaffint x xcall_list =
  let (lblin, lblout) = new_var_code x in
  output_string !oc ("
     latexout(\"beginning\");
     goto " ^  lblin ^ ";\n"
     ^ lblout ^ ":
     latexout(\"printing\");
     printf(\"%li\",s_top(vstack));
     s_pop(vstack);\n");
  xcall_list := ((lblin, lblout), 0) :: !xcall_list

and

codeaffbool x xcall_list =
  let (lblin, lblout) = new_var_code x in
  output_string !oc ("
     latexout(\"beginning\");
     goto " ^  lblin ^ ";\n"
     ^ lblout ^ ":
     latexout(\"printing\");
     if (s_top(vstack))
        printf(\"tt\");
     else
        printf(\"ff\");
     s_pop(vstack);\n");
  xcall_list := ((lblin, lblout), 0) :: !xcall_list

and

codeafflist x t xcall_list =
  let (lblin, lblout) = new_var_code x in
  xcall_list := ((lblin, lblout), 0) :: !xcall_list;
  output_string !oc ("
     printf(\"[\");
     do {
          _DSTACKI
          latexout(\"beginning\");
          goto " ^ lblin ^ ";\n"
          ^ lblout ^ ":
          if (s_top(vstack))
             break;
          s_pop(vstack);
          _DSTACKH");
  codeaff x t xcall_list;
  output_string !oc "
          _DSTACKL
     } while (1);
     s_pop(vstack);
     _CUTDSTACK
     printf(\"]\");\n"

and

codeaffdumb ()=
  output_string !oc "
     latexout(\"printing\");
     printf(\"<function>\");\n"

and codeaff x ty xcall_list = 
               match (Typing.instanvar ty).type_desc with
                   TConst "int" -> codeaffint x xcall_list
                 | TConst "bool" -> codeaffbool x xcall_list
                 | TList t -> codeafflist x t xcall_list
                 | _ -> codeaffdumb()
;;

let codemult (l1, l1out) (l2, l2out) (l3, l3out) =
  output_string !oc (l1 ^ ":
     R0 <<= 1;
     latexout(\"p\");
     goto " ^ l3out ^ ";\n"
     ^ l2 ^ ":
     R0 = (R0 << 1) + 1;
     latexout(\"q\");
     goto " ^ l3out ^ ";\n"
     ^ l3 ^":
     if (R0 & 1) {
        R0 >>= 1;
        latexout(\"q*\");
        goto " ^ l2out ^ ";
     } else {
        R0 >>= 1;
        latexout(\"p*\");
        goto " ^ l1out ^ ";
     };\n\n")
;;

let codeweakmult (l2, l2out) (l3, l3out) =
  output_string !oc (l2 ^ ":
     R0 = (R0 << 1) + 1;
     latexout(\"q\");
     goto " ^ l3out ^ ";\n"
     ^ l3 ^":
     R0 >>= 1;
     latexout(\"q*\");
     goto " ^ l2out ^ ";\n\n")
;;

let codecond (l1, l1out) (l2, l2out) (l3, l3out) (l4, l4out) =
  output_string !oc (l1 ^ ":
     goto " ^ l2out ^ ";\n"
     ^ l3 ^ ":\n"
     ^ l4 ^ ":
     goto " ^ l1out ^ ";\n"
     ^ l2 ^ ":
     if (s_top(vstack)) {
        s_pop(vstack);
        latexout(\"cond\");
        goto " ^ l3out ^ ";
     } else {
        s_pop(vstack);
        latexout(\"cond\");
        goto " ^ l4out ^ ";
     }\n\n")
;;

let codederelic (l1, l1out) (l2, l2out) =
  output_string !oc (l1 ^ ":
     _DER
     goto " ^ l2out ^ ";\n"
     ^ l2 ^ ":
     _DERREV
     goto " ^ l1out ^ ";\n\n")
;;

let codecontract v (l1, l1out) (l2, l2out) (l3, l3out) =
  output_string !oc (l1 ^ ":
     { cont_t *tmp;
       h_malloc(hcont,tmp);
       tmp->d = 0;
       tmp->next = ep->left->content;
       ep->left->content = tmp;
     };
     latexout(\"r\");
     goto " ^ l3out ^ ";\n"
     ^ l2 ^ ":
     { cont_t *tmp;
       h_malloc(hcont,tmp);
       tmp->d = 1;
       tmp->next = ep->left->content;
       ep->left->content = tmp;
     };
     latexout(\"s\");
     goto " ^ l3out ^ ";\n"
     ^ l3 ^ ":
     { cont_t *tmp = ep->left->content;
       ep->left->content = tmp->next;
       if (tmp->d) {
          h_free(hcont,tmp);
          latexout(\"s*\");
          goto " ^ l2out ^ ";
       } else {
          h_free(hcont,tmp);
          latexout(\"r*\");
          goto " ^ l1out ^ ";
       };
     };\n\n")
;;

let codeofcourse (l1, l1out) (l2, l2out) =
  output_string !oc (l1 ^ ":
     ep = ep->right;
     latexout(\"!\");
     goto " ^ l2out ^ ";\n"
     ^ l2 ^":
     ep = ep->prev;
     latexout(\"!*\");
     goto " ^ l1out ^ ";\n\n")
;;

let codequestion (l1, l1out) (l2, l2out) =
  output_string !oc (l1 ^ ":
     _QUESTION
     goto " ^ l2out ^ ";\n"
     ^ l2 ^ ":
     _QUESTIONREV
     goto " ^ l1out ^ ";\n\n")
;;

let codenil (l1, l1out) =
  output_string !oc (l1 ^ ":
     if (dstack->d == 'i') {
        s_push(vstack,tt);
        latexout(\"i*-tt-i\");
        goto " ^ l1out ^ ";
     } else {
        printf(\"Bad access to nil node\\n\");
        latexend();
        latexendall();
        exit(0);
     }\n\n")
;;

let codelistask c (l1, l1out) (l2, l2out) =
  output_string !oc (l1 ^ ":
     { dat_ac_t *tmp;
       h_malloc(hdat,tmp);
       tmp->d = '" ^ c ^ "';
       tmp->next = dstack;
       dstack = tmp;
       latexout(\"" ^ c ^ "\");
       goto " ^ l2out ^ ";
     }\n"
     ^ l2 ^ ":
     { dat_ac_t *tmp = dstack;
       dstack = dstack->next;
       h_free(hdat,tmp);
       latexout(\"" ^ c ^ "*\");
       goto " ^ l1out ^ ";
     }\n\n")
;;

let codecst s (l, lout) =
  output_string !oc (l ^ ":
    s_push(vstack, " ^ s ^ ");
    latexout(\"" ^ s ^ "\");
    goto " ^ lout ^ ";\n\n")
;;
       
let codeinternal name (l1, l1out) (l2, l2out) =
  output_string !oc (l1 ^ ":
    goto " ^ l2out ^ ";\n"
    ^ l2 ^ ":
    s_top(vstack) " ^
      (match name with
          "iszero" | "not" -> "= (s_top(vstack) == 0)"
        | "succ" -> "++"
        | "pred" -> "--"
        | _ -> failwith "Internal error : undefined language identifier")
    ^ ";
    latexout(\"" ^ name ^"\");
    goto " ^ l1out ^ ";\n\n")
;;

let codeop name (l1, l1out) (l2, l2out) (l3, l3out) =
match name with
   "::" ->
  output_string !oc (l1 ^ ":
     {
       if (dstack->d == 'i') {
         s_push(vstack,ff);
         latexout(\"i*-ff-i\");
         goto " ^ l1out ^ ";
       } else if (dstack->d == 'h') {
         dat_ac_t *tmp = dstack;
         dstack = dstack->next;
         h_free(hdat,tmp);
         latexout(\"h*\");
         goto " ^ l2out ^ ";
       } else {
         dat_ac_t *tmp = dstack;
         dstack = dstack->next;
         h_free(hdat,tmp);
         latexout(\"l*\");
         goto " ^ l3out ^ ";
       }
     }\n"
    ^ l2 ^ ":
     { dat_ac_t *tmp;
       h_malloc(hdat,tmp);
       tmp->d = 'h';
       tmp->next = dstack;
       dstack = tmp;
       latexout(\"h\");
       goto " ^ l1out ^ ";
     }\n"
    ^ l3 ^ ":
     { dat_ac_t *tmp;
       h_malloc(hdat,tmp);
       tmp->d = 'l';
       tmp->next = dstack;
       dstack = tmp;
       latexout(\"l\");
       goto " ^ l1out ^ ";
     }\n\n")
  | "or" ->
  output_string !oc (l1 ^ ":
    goto " ^ l2out ^ ";\n"
    ^ l2 ^ ":
    if (s_top(vstack) == 0) {
      s_pop(vstack);
      latexout(\"or\");
      goto " ^ l3out ^ ";
    } else
      goto " ^ l1out ^ ";\n"
    ^ l3 ^ ":
    goto " ^ l1out ^ ";\n\n")
  | "and" ->
  output_string !oc (l1 ^ ":
    goto " ^ l2out ^ ";\n"
    ^ l2 ^ ":
    if (s_top(vstack)) {
      s_pop(vstack);
      latexout(\"and\");
      goto " ^ l3out ^ ";
    } else
      goto " ^ l1out ^ ";\n"
    ^ l3 ^ ":
    goto " ^ l1out ^ ";\n\n")
  | _ ->
  output_string !oc (l1 ^ ":
    goto " ^ l2out ^ ";\n"
    ^ l2 ^ ":
    goto " ^ l3out ^ ";\n"
    ^ l3 ^ ":
    { signed long tmp = s_top(vstack);
      s_pop(vstack);
      s_top(vstack) = (s_top(vstack) " ^ (match name with 
        "=" -> "=="
      | "<>" | "xor" -> "!="
      | _ -> name) 
          ^ " tmp);
    }
    latexout(\"" ^ (match name with
         "%" -> "\\\\%"
       | "<=" -> "$<=$"
       | "<" -> "$<$"
       | "<>" -> "$<>$"
       | _    ->  name) ^ "\");
    goto " ^ l1out ^ ";\n\n")
;;

let codewire (l1, l1out) (l2, l2out) =
  output_string !oc (l1 ^ ":
    goto " ^ l2out ^ ";\n"
    ^ l2 ^ ":
    goto " ^ l1out ^ ";\n\n")
;;

let codeerr (l,_) mess =
  output_string !oc (l ^ ":
    printf(\"" ^ mess ^ "\\n\");
    latexend();
    latexendall();
    exit(1);\n")
;;
  
let globaldecl = ref ([] : (string * sexpr) list);;
 
(* ident, ref [((in_label, out_label), box_level)] *)
(* special identifiers
   are coded by their corresponding node on EACH call, instead of making
   contractions and box doors, if they are immediately applied to arguments.
   Else, they are coded according to expressions given in globaldecl *)

let globalsym = ref ([] : (string * ((string * string) * int) list ref ) list)
;;

let box_level = ref 0;;

let swap (a, b) = (b, a);;

(* Caution ! On each call to varaccess, we must react specially if
   call_list is empty ! *)

let rec pairaccesses v = 
  function (((lbl1,lev1)::(lbl2,lev2)::l) as l1) -> 
        if (lev1 = !box_level) & (lev2 = !box_level) then
        begin 
           let lbl3 = new_label() in
              codecontract v lbl1 lbl2 (swap lbl3);
              ((lbl3,lev2):: pairaccesses v l)
        end
        else l1
   | l -> l
;;

exception Empty_call_list;;

let rec varaccess v =
   function (((lbl1,lev1)::(lbl2,lev2)::l) as l1) ->
        if (lev1 = !box_level) & (lev2 = !box_level) then
           varaccess v (pairaccesses v l1)
        else if lev1 = !box_level then
           lbl1
        else 
           raise Empty_call_list
    | [(lbl1,lev1)] -> if lev1 = !box_level then
          lbl1
        else
           raise Empty_call_list
    | [] -> raise Empty_call_list
;; 

let rec clear_this_level =
   function [] -> []
    | (((lbl,lev)::l) as l1) -> 
         if lev = !box_level then 
            clear_this_level l
         else 
            l1
;;

let rec question_this_level =
   function [] -> ()
    | ((x,xcall_list)::l) -> 
        (try
           let lbl = new_label() in
           codequestion (varaccess x !xcall_list) (swap lbl);
           xcall_list := ((lbl,!box_level-1) :: (clear_this_level !xcall_list))
        with
           Empty_call_list -> () );
        question_this_level l
;;

let rec code_expr symtable lbl e0 = match e0 with
  Func(x,e) -> let xcall_list = ref []
               and l=new_label()
               and l2=new_label() in
                  incr box_level;
                  code_expr ((x, xcall_list) :: symtable) l2 e;
                  codeofcourse lbl (swap l);
                  (try
                     codemult (varaccess x !xcall_list) (swap l2) l
                  with
                     Empty_call_list -> codeweakmult (swap l2) l);
                  question_this_level symtable;
                  decr box_level
| F(x,e2) -> begin
             match x with
               "hd" | "tl" | "isnil" ->
                 let l = new_label() in
                    code_expr symtable l e2;
                    codelistask (match x with 
                      "hd" -> "h"
                    | "tl" -> "l"
                    | "isnil" -> "i"
                    | _ -> failwith "internal error: unknown list op") 
                      lbl (swap l)
             | _ ->
                 let l = new_label() in
                    code_expr symtable l e2;
                    codeinternal x lbl (swap l)
            end
| Rec(phi,e) ->
     let l1 = new_label()
     and phicall_list = ref []
     in code_expr ((phi,phicall_list)::symtable) l1 e;
        (try
            codecontract phi lbl (varaccess phi !phicall_list) (swap l1)
         with
            Empty_call_list -> 
                print_endline "Internal error : useless rec \
                     should have been removed by optimization";
                codewire lbl (swap l1)) 
| Op(x, e3, e2) ->
     let l2 = new_label()
     and l3 = new_label()
     in code_expr symtable l2 e2;
        code_expr symtable l3 e3;
        codeop x lbl (swap l3) (swap l2)
| Cond(be,e3,e2) ->
     let l1 = new_label()
     and l2 = new_label()
     and l3 = new_label()
     in code_expr symtable l1 be;
        code_expr symtable l2 e2;
        code_expr symtable l3 e3;
        codecond lbl (swap l1) (swap l3) (swap l2)
| Apply(e1,e2) ->
     let l1=new_label() 
     and l2=new_label() 
     and l3=new_label() in
         code_expr symtable l1 e1;
         code_expr symtable l2 e2;
         codederelic l3 (swap l1);
         codemult (swap l2) lbl (swap l3)
| Var(x) -> begin
             match x with
                 "tt" | "ff" -> codecst x lbl
               | "[]" -> codenil lbl
               | "_match_error" -> codeerr lbl "Pattern matching failed"
               | _ -> try
                        let xcall_list = List.assoc x symtable in
                         xcall_list := (lbl,!box_level) :: !xcall_list
                      with 
                        Not_found -> raise (Error(e0, "Variable "^x^
                            " not declared"))
            end
| Int n -> codecst (string_of_int n) lbl
| Affectin(x,e1,e2) -> let xcall_list = ref []
                       in code_expr ((x, xcall_list)::symtable) lbl e2;
                          try
                            code_expr symtable (varaccess x !xcall_list) e1 
                          with Empty_call_list -> ()
;;

let rec record_prog = function 
  [] ->  globaldecl := [];
         globalsym := []
 | (x, ty, e, _)::syml ->
  record_prog syml; 
  let xcall_list = ref [] in
     globaldecl := (x, e) :: !globaldecl;
     globalsym := (x, xcall_list) :: !globalsym;
     if (String.length x > 5) & (String.sub x 0 5 = "_expr") then
     begin
        output_string !oc "
      latexstart();
      fputs(\"";
        print_prog !oc (Expr e);
        output_string !oc "Result : \", stdout);\n";
        if !repeat > 1 then
           output_string !oc ("
      for (r=1; r<=" ^ (string_of_int !repeat) ^ "; r++) {");
        codeaff x ty xcall_list;
        if !repeat > 1 then
           output_string !oc "
      printf(\" - \");
      }";
        output_string !oc "
      printf(\"\\n\");
      latexend();\n\n"
     end
;;

let rec code_prog = function
       ([],[]) -> ()
    |  ((x, xcall_list) :: syml, (y,e)::decll) -> 
         (if x = y then
            try
               code_expr syml (varaccess x !xcall_list) e
            with 
               Empty_call_list -> ()
             | Error(e0, msg) ->
                 print_underline (Affect(x,e)) e0;
                 print_endline msg;
                 raise Bad_end
         else
            failwith "Internal error : Bad declarations");
         code_prog (syml,decll)
   | _ -> failwith "Internal error : Bad declarations"
;; 

let code symtable =
  label_count := 0;
  if !latex then
     output_string !oc "#define latex\n\n";
  if !count then
     output_string !oc "#define count\n\n";
  output_string !oc ("/* This program was generated by sflcomp */

#include \"" ^ !Config.macrodir ^ "sflconfig.c\"
#include \"" ^ !Config.macrodir ^ "nonoptgnammacros.c\"

ulng r;

int main()
{");
  if !meml > 0 then
     output_string !oc ("
  putlimit(RLIMIT_DATA, " ^ (string_of_int (!meml * 1024)) ^ ");");
  if !timel > 0 then
     output_string !oc ("
  putlimit(RLIMIT_CPU, " ^ (string_of_int !timel) ^ ");");
  output_string !oc ("
     tree1.prev = tree2.prev = &tree3;
     latexstartall();
     printf(\"" ^ (machinename !mach) ^ " - SFL program running...\\n\");
\n");
  record_prog symtable;
  output_string !oc "
     latexendall();
     return 0;\n\n";
  code_prog (!globalsym, !globaldecl);
  output_string !oc "};\n"
;;

