/* SFL parser definition */

%{
open Globals;;
%}

/* Tokens */

%token <string> IDENT
%token <int> INT
%token <string> STRING
%token SHARP
%token EOF
%token LET
%token REC
%token IN
%token IF
%token THEN
%token ELSE
%token RBRACKET
%token LBRACKET
%token FUN
%token ARROW
%token COLON
%token QUOTE
%token COMMA
%token EQUAL
%token LESSEQUAL
%token MOREEQUAL
%token DIFF
%token UNDERSCORE
%token LESS
%token MORE
%token LPAREN
%token RPAREN
%token SEMICOLON
%token SEMISEMI
%token PLUS
%token MINUS
%token TIMES
%token DIVIDE
%token REMAINDER
%token CONS
%token BAR
%token AND
%token OR
%token XOR
%token <string> PREDEF
%token MATCH
%token WITH

/* Priorities and asociativities */

%left EQUAL LESSEQUAL LESS MOREEQUAL MORE DIFF
%left PLUS MINUS OR XOR
%left TIMES DIVIDE REMAINDER AND

/* Entry point */

%start prog
%type <Globals.zprogelem> prog
%start top
%type <Globals.top_phrase> top

%%

top	: SHARP IDENT param SEMISEMI
		{ Command($2,$3) }
	| prog
		{ Prog $1 }
	;

param	: IDENT param
		{ (PId $1)::$2 }
	| STRING param
		{ (PStr $1)::$2 }
	| INT param
		{ (PInt $1)::$2 }
	| 
		{ [] }
	;

prog	: LET IDENT EQUAL texpr SEMISEMI
		{ ZAffect($2,$4) } 
	| LET REC IDENT EQUAL texpr SEMISEMI
                { ZAffect($3,make_expr (ZRec($3,$5))) }
	| texpr SEMISEMI
		{ ZExpr($1) }
	| EOF
		{ raise End_of_file }
	;

texpr	: expr COLON typeexpr
		{ make_expr (ZType($1,$3)) }
	| expr
		{ $1 }
	;

expr	: sexpr
                { $1 }
        | simpleexprlist sexpr
                { make_expr (ZApply($1,$2)) }
	;

sexpr	: simpleexpr
		{ $1 }
	| FUN fundecl
		{ make_expr (ZFunc($2)) }
	| MATCH expr WITH fundecl
		{ make_expr (ZMatch($2,$4)) }
	| LET IDENT EQUAL expr IN expr
		{ make_expr (ZAffectin($2,$4,$6)) }
	| LET REC IDENT EQUAL expr IN expr
                { make_expr (ZAffectin($3,make_expr (ZRec($3,$5)),$7)) }
        | IF expr THEN expr ELSE expr
                { make_expr (ZCond($2,$4,$6)) }
	| simpleexpr CONS expr
		{ make_op "::" $1 $3 }
	;

simpleexpr
	: LPAREN texpr RPAREN
		{ $2 }
	| IDENT
		{ make_expr (ZVar $1) }
	| PREDEF
		{ make_expr (ZVar $1) }
	| INT
		{ make_expr (ZInt $1) }
        | simpleexpr PLUS simpleexpr
		{ make_op "+" $1 $3 }
        | simpleexpr MINUS simpleexpr
		{ make_op "-" $1 $3 }
        | simpleexpr TIMES simpleexpr
		{ make_op "*" $1 $3 }
        | simpleexpr DIVIDE simpleexpr
		{ make_op "/" $1 $3 }
        | simpleexpr REMAINDER simpleexpr
		{ make_op "%" $1 $3 }
        | simpleexpr EQUAL simpleexpr
		{ make_op "=" $1 $3 }
        | simpleexpr LESSEQUAL simpleexpr
		{ make_op "<=" $1 $3 }
        | simpleexpr LESS simpleexpr
		{ make_op "<" $1 $3 }
        | simpleexpr DIFF simpleexpr
		{ make_op "<>" $1 $3 }
        | simpleexpr AND simpleexpr
		{ make_op "and" $1 $3 }
        | simpleexpr OR simpleexpr
		{ make_op "or" $1 $3 }
        | simpleexpr XOR simpleexpr
		{ make_op "xor" $1 $3 }
        | simpleexpr MOREEQUAL simpleexpr
		{ make_op "<=" $3 $1 }
        | simpleexpr MORE simpleexpr
		{ make_op "<" $3 $1 }
        | LBRACKET exprlist RBRACKET
                { $2 }
	| LBRACKET RBRACKET
		{ make_expr (ZVar "[]") }
	;

simpleexprlist
	: simpleexprlist simpleexpr
		{ make_expr (ZApply($1,$2)) }
	| simpleexpr
		{ $1 }
	;

exprlist
	: texpr SEMICOLON exprlist
		{ make_op "::" $1 $3 }
	| texpr
		{ make_op "::" $1 (make_expr (ZVar "[]")) }
	;

fundecl	: pattern ARROW expr BAR fundecl
		{ ($1,$3)::$5 }
	| pattern ARROW expr
		{ [$1,$3] }
	;

pattern : patternnc
		{ $1 }
	| patternnc CONS pattern
		{ make_pat (Cons($1,$3)) }
	;

patternnc
	: LBRACKET RBRACKET
		{ make_pat Nil }
	| LBRACKET patlist RBRACKET
                { $2 }
        | IDENT
		{ make_pat (Id $1) }
	| INT
		{ make_pat (In $1) }
	| UNDERSCORE
		{ make_pat Any }
        | LPAREN pattern RPAREN
		{ $2 }
	;

patlist	: pattern SEMICOLON patlist
		{ make_pat (Cons($1,$3)) }
	| pattern
		{ make_pat (Cons($1,make_pat Nil)) }
	;

typeexpr
	: stypeexpr ARROW typeexpr
		{ make_type (ZTypeArrow($1,$3)) }
	| stypeexpr
		{ $1 }
	;

stypeexpr
	: QUOTE IDENT
		{ make_type (ZTypeVar $2) }
	| stypeexpr IDENT
		{ make_type (ZTypeConstr($1,$2)) }
	| IDENT
		{ make_type (ZTypeConst $1) }
	| LPAREN typeexpr RPAREN
		{ $2 }
	;
