/* Macro include file for the call by value machine */

/* Possible contents of ep->l */
#define _L 1
#define _v 2  /* Value */

#define NIL (env_t *) NULL

typedef union {
        void *addr;       /* if l = _L or 0 */
        signed long val;  /* if l = _v */
        /* nothing */     /* if l = _h, _l or _i */
} v_t;
 
typedef struct env_s {
        struct env_s *next, *cont;
        v_t v;
        bool l;
} env_t;

heap_t henv = h_init(env_t);
stack_t dstack = s_init;
env_t *ep,*context;

#ifdef latex

void printenv(env_t *env)
{
  if (env == NIL) 
     fprintf(f,"\\carre");
  else {
     fprintf(f, "$<$");
     switch (env->l) {
       case _v : fprintf(f,"%li",env->v.val);
                 break;
       case _h : fprintf(f, "h");
                 break; 
       case _l : fprintf(f, "l");
                 break; 
       case _i : fprintf(f, "i");
                 break; 
       case _L : printenv(env->cont);
                 fprintf(f,":L");
                 break;
       case 0 :printenv(env->cont);
               break;
       default : fprintf(f,"Bad");
       }
     fprintf(f, ",");
     printenv(env->next);
     fprintf(f, "$>$");
   }
};

char *latexpreamble()
{
   return "\\tvi\\quad#\\hfill\\quad&\n\
\\tvi\\hfill\\quad#\\hfill\\quad&\n\
\\tvi\\hfill\\quad#\\hfill\\quad\\tvi\\cr\n\
\\noalign{\\hrule}\n\
Instruc.&   Environment& Stack\\cr\n";
};

void latexprintline(char *s)
{
  fprintf(f,"{\\it %s}& ",s);
  printenv(context);
  fprintf(f,"& ");
  printenv(ep);
  fprintf(f, "\\cr\n");
};

#endif

#define _INIT(lbl) \
     context = NIL;\
     initstack(&&lbl);\
     latexout("beginning");

#define _DERNS(l1out) \
     { env_t *a;\
       h_malloc(henv,a);\
       a->next = ep;\
       ep = a;\
       ep->l = _L;\
       ep->cont = NIL;\
       ep->v.addr = &&l1out;\
     }

#define _DERREV \
     { env_t *tmp = ep;\
       freetree(ep->cont);\
       ep = ep->next;\
       h_free(henv,tmp);\
     }

#define _PUTVAL(x,mess) \
    { void *add = ep->v.addr;\
      freetree(context);\
      context = ep;\
      ep = ep->next;\
      context->next = context->cont;\
      context->cont = NIL;\
      context->l = _v;\
      context->v.val = x;\
      latexout(mess);\
      goto *add;\
    }

#define _POPCONTEXTVALUE \
    { env_t *tmp = context;\
      context = context->next;\
      h_free(henv,tmp);\
    }

#define _ENVTOSTACK \
    { env_t *tmp = context;				\
      context = context->next;				\
      tmp->next = ep;					\
      ep = tmp;	           				\
    }

#define _SWAPCONT \
    { env_t *tmpcont = ep->cont;\
      ep->cont = context;\
      context = tmpcont;\
    }

#define _SWAP2 \
    { env_t *tmp = ep;\
      ep = ep->next;\
      tmp->next = tmp->cont;\
      tmp->cont = context;\
      context = tmp;\
    }

void initstack(void *add)
{
   env_t *tmp;
   stack_t dstacktmp = dstack;
   ulng i;
   
   h_malloc(henv,ep);
   ep->v.addr = add;
   ep->cont = NIL;
   ep->l = 0;
   ep->next = NIL;
   while (! s_empty(dstacktmp)) {
      h_malloc(henv,tmp);
      tmp->next = ep;
      tmp->cont = NIL;
      tmp->l = ((s_top(dstacktmp) & 1) == 0 ? _h : _i);
      ep = tmp;
      for (i = 1; i<=s_top(dstacktmp) / 2; i++) {
         h_malloc(henv,tmp);
         tmp->next = ep;
         tmp->cont = NIL;
         tmp->l = _l;
         ep = tmp;
      }
      s_next(dstacktmp);
   }
};

env_t *copytree(env_t *p)
{
  if (p != NIL) {
    env_t *a;
    h_malloc(henv,a);
#ifdef count
    nbr[2]++;
#endif
    a->next = copytree(p->next);
    a->cont = copytree(p->cont);
    a->v = p->v;
    a->l = p->l;
    return a;
  }
  else return NIL;
};

void freetree(env_t *p)
{
  if (p != NIL) {
     freetree(p->next);
     freetree(p->cont);
     h_free(henv,p);
  }
};
