/* Macro include file for the call by need machine */

/* Possible contents of env_t.l */
#define _UPD 1
#define _v 2  /* Value */
#define _cv 6 /* Complex values. optimisation to avoid update on
                 each access to functions or lists */

#define NIL (env_t *) NULL

typedef union {
        signed long val;        /* if l = _v */
        void *addr;             /* if l = 0 or _cv */
        /* nothing */           /* if l = _h, _l, _i or _UPD */
} v_t;

typedef struct env_s {
        struct env_s *next, *cont;
        v_t v;
        unsigned short nbrcopy;
        bool l;
} env_t;

heap_t henv = h_init(env_t);
stack_t dstack = s_init;
env_t *ep,*context;

#ifdef latex

void printenv(env_t *env)
{
  if (env == NIL) 
     fprintf(f,"\\carre");
  else {
     switch (env->l) {
       case _v : fprintf(f,"%li",env->v.val);
                 break;
       case _h : fprintf(f, "h");
                 break; 
       case _l : fprintf(f, "l");
                 break; 
       case _i : fprintf(f, "i");
                 break; 
       case _UPD :fprintf(f,"UPD");
                  break;
       case _cv :
       case 0 : fprintf(f, "N");
       }
     fprintf(f, ":");
     printenv(env->next);
   }
};

char *latexpreamble()
{
   return "\\tvi\\quad#\\hfill\\quad&\n\
\\tvi\\hfill\\quad#\\hfill\\quad&\n\
\\tvi\\hfill\\quad#\\hfill\\quad\\tvi\\cr\n\
\\noalign{\\hrule}\n\
Instruc.&   Environment& Stack\\cr\n";
};

void latexprintline(char *s)
{
  fprintf(f,"{\\it %s}& ",s);
  printenv(context);
  fprintf(f,"& ");
  printenv(ep);
  fprintf(f, "\\cr\n");
};

#endif

#define _INIT(lbl) \
    context = NIL;\
    initstack(&&lbl);\
    latexout("beginning");

#define _DER(l1out) \
     { env_t *a;\
       h_malloc(henv,a);\
       a->next = ep;\
       ep = a;\
       ep->nbrcopy = 1;\
       if (context != NIL) context->nbrcopy++;\
       ep->cont = context;\
       ep->l = 0;\
       ep->v.addr = &&l1out;\
     }

#define _LETDECL(l1out) \
     { env_t *a;\
       h_malloc(henv,a);\
       a->next = context;\
       a->nbrcopy = 1;\
       if (context != NIL) context->nbrcopy++;\
       a->cont = context;\
       a->l = 0;\
       a->v.addr = &&l1out;\
       context = a;\
     }\
     latexout("let decl");

#define _PUTUPD(updaddr) \
     { env_t *a;\
       h_malloc(henv,a);\
       a->next = ep;\
       ep = a;\
       ep->nbrcopy = 1;\
       ep->cont = updaddr;\
       updaddr->nbrcopy++;\
       ep->l = _UPD;\
     }

/* We don't free the 'cont' tree, because either it is NIL or
   it has been reused somewhere else */
#define _DERREV \
     { env_t *tmp = ep;\
       ep = ep->next;\
       h_free(henv,tmp);\
     }


/* We don't need to make the updating of a cell, if it will be
   discarded immediately after (ep->cont->nbrcopy == 1) */
#define _UPDATE(x) \
    while (ep->l == _UPD)\
    { \
      if (ep->cont->nbrcopy > 1) {\
        freetree(ep->cont->cont);\
        ep->cont->v.addr = x;\
        ep->cont->l = _cv;\
        if (context != NIL) context->nbrcopy++;\
        ep->cont->cont = context;\
      }\
      freetree(ep->cont);\
      _DERREV\
    }

#define _PUTVAL(x,mess) \
  { while (ep->l == _UPD)\
    { \
      if (ep->cont->nbrcopy > 1) {\
        freetree(ep->cont->cont);\
        ep->cont->v.val = x;\
        ep->cont->l = _v;\
        ep->cont->cont = NIL;\
      }\
      freetree(ep->cont);\
      _DERREV\
    }\
    freetree(context);\
    context = ep->cont;\
    { void *add = ep->v.addr;\
      _DERREV\
      latexout(mess);\
      goto *add;\
    }\
  }

#define _ASKVAL(addupd,lblret) \
    { env_t *tmp, *tmp2;\
      h_malloc(henv,tmp);\
      tmp2 = tmp;\
      tmp->cont = addupd;\
      addupd->nbrcopy++;\
      tmp->nbrcopy = 1;\
      tmp->l = _UPD;\
      h_malloc(henv,tmp->next);\
      tmp = tmp->next;\
      tmp->v.addr = &&lblret;\
      tmp->nbrcopy = 1;\
      if (context != NIL) context->nbrcopy++;\
      tmp->cont = context;\
      tmp->l = 0;\
      tmp->next = ep;\
      ep = tmp2;\
    }

#define _ASKVALPLACE(lblret) \
    { env_t *tmp, *tmp2;\
      h_malloc(henv,tmp);\
      tmp2 = tmp;\
      tmp->l = _UPD;\
      tmp->nbrcopy = 1;\
      h_malloc(henv,tmp->next);\
      tmp = tmp->next;\
      tmp->v.addr = &&lblret;\
      tmp->nbrcopy = 1;\
      if (context != NIL) context->nbrcopy++;\
      tmp->cont = context;\
      tmp->l = 0;\
      h_malloc(henv,tmp->next);\
      tmp = tmp->next;\
      tmp->v.val = 0;\
      tmp->l = _v;\
      tmp->cont = NIL;\
      tmp->nbrcopy = 2;\
      tmp2->cont = tmp;\
      tmp->next = ep;\
      ep = tmp2;\
    }

/* Sometimes, in codeinternal or in codeop, the environment need not
   be saved because when don't need it to compute succ x... 
   This enables to free environments sooner */

#define _ASKVALNS(addupd,lblret) \
    { env_t *tmp, *tmp2;\
      h_malloc(henv,tmp);\
      tmp2 = tmp;\
      tmp->cont = addupd;\
      addupd->nbrcopy++;\
      tmp->nbrcopy = 1;\
      tmp->l = _UPD;\
      h_malloc(henv,tmp->next);\
      tmp = tmp->next;\
      tmp->v.addr = &&lblret;\
      tmp->nbrcopy = 1;\
      tmp->cont = NIL;\
      tmp->l = 0;\
      tmp->next = ep;\
      ep = tmp2;\
    }

#define _ASKVALPLACENS(lblret) \
    { env_t *tmp, *tmp2;\
      h_malloc(henv,tmp);\
      tmp2 = tmp;\
      tmp->l = _UPD;\
      tmp->nbrcopy = 1;\
      h_malloc(henv,tmp->next);\
      tmp = tmp->next;\
      tmp->v.addr = &&lblret;\
      tmp->nbrcopy = 1;\
      tmp->cont = NIL;\
      tmp->l = 0;\
      h_malloc(henv,tmp->next);\
      tmp = tmp->next;\
      tmp->v.val = 0;\
      tmp->l = _v;\
      tmp->cont = NIL;\
      tmp->nbrcopy = 2;\
      tmp2->cont = tmp;\
      tmp->next = ep;\
      ep = tmp2;\
    }

void initstack(void *add)
{
   env_t *tmp;
   stack_t dstacktmp = dstack;
   ulng i;
   
   h_malloc(henv,tmp);
   ep = tmp;
   tmp->l = _UPD;
   tmp->nbrcopy = 1;
   h_malloc(henv,tmp->next);
   tmp = tmp->next;
   tmp->v.addr = add;
   tmp->nbrcopy = 1;
   tmp->cont = context;
   tmp->l = 0;
   h_malloc(henv,tmp->next);
   tmp = tmp->next;
   tmp->v.val = 0;
   tmp->cont = NIL;
   tmp->nbrcopy = 2;
   tmp->l = _v;
   tmp->next = NIL;\
   ep->cont = tmp;\
   while (! s_empty(dstacktmp)) {
      h_malloc(henv,tmp);
      tmp->next = ep;
      tmp->l = ((s_top(dstacktmp) & 1) == 0 ? _h : _i);
      tmp->cont = NIL;
      tmp->nbrcopy = 1;
      ep = tmp;
      for (i = 1; i<=s_top(dstacktmp) / 2; i++) {
         h_malloc(henv,tmp);
         tmp->next = ep;
         tmp->l = _l;
         tmp->cont = NIL;
         tmp->nbrcopy = 1;
         ep = tmp;
      }
      s_next(dstacktmp);
   }
};

void freetree(env_t *p)
{
  if (p != NIL) {
    p->nbrcopy--;
    if (p->nbrcopy == 0) {
       freetree(p->next);
       freetree(p->cont);
       h_free(henv,p);
    }
  }
};
