/* Macro include file */
/* Common part between GOI and question stack machines */

#define NIL (env_t *) NULL
#define CNIL (cont_t *) NULL
     
typedef struct cont {
        void *d;
        struct cont *next;
} cont_t;
 
typedef struct env {
        struct env *prev, *right, *left;
        cont_t *content;
} env_t;

ulng R0 = 1;
stack_t vstack = s_init;
stack_t dstack = s_init;
stack_t dstack2 = s_init;
heap_t henv = h_init(env_t);
heap_t hcont = h_init(cont_t);
env_t *ep, *context;

#ifdef latex

void printenv(env_t *env)
{
  if (env == NIL) 
     fprintf(f,"\\carre");
  else {
     cont_t *a = env->content;
     while (a != CNIL)
     {
        if (a->d != NULL) 
           fprintf(f, "R:");
        else
           fprintf(f, "L:");
        a = a->next;
     }
     if (env == ep) 
        fprintf(f,"{\\it ep:}");
     fprintf(f, "$<$");
     printenv(env->left);
     fprintf(f, ",");
     printenv(env->right);
     fprintf(f, "$>$");
   }
};

char *latexpreamble()
{
  return "\\tvi\\quad#\\hfill\\quad&\n\
\\tvi\\hfill\\quad#\\quad&\n\
\\tvi\\hfill\\quad#\\hfill\\quad&\n\
\\tvi\\hfill\\quad#\\quad&\n\
\\tvi\\hfill\\quad#\\quad\\tvi\\cr\n\
\\noalign{\\hrule}\n\
Instruction&  Mult&   Environment& Data access& Data\\cr\n";
};

void latexprintline(char *s)
{
  ulng a = R0;
  stack_t tmp = vstack;
  stack_t d = dstack2;

  fprintf(f, "{\\it %s}& {\\it ",s);
  while (a > 1)
  {
     if (a & 1) 
        fprintf(f, "r:");
     else
        fprintf(f, "l:");
     a >>= 1;
   };
  fprintf(f, "}\\carre & ");
  printenv(context);
  fprintf(f, "& ");
  while (! s_empty(d))
  { ulng i;

    for (i=1; i<=s_top(d) / 2; i++)
       fprintf(f, "l:");
    if ((s_top(d) & 1) == 0)
       fprintf(f, "h:");
    else
       fprintf(f, "i:");
    s_next(d);
  }
  fprintf(f, "\\carre & ");  
  while (! s_empty(tmp))
  {
    fprintf(f, "%li:", s_top(tmp));
    s_next(tmp);
  }
  fprintf(f, "\\carre\\cr\n");
};

#endif

#define _INIT \
     copydatac();\
     latexout("beginning");

#define _DER \
     { env_t *a;\
       h_malloc(henv,a);\
       a->right = ep->right;\
       if (a->right != NIL) a->right->prev = a;\
       a->left = ep->left;\
       if (a->left != NIL) a->left->prev = a;\
       a->content = ep->content;\
       a->prev = ep;\
       ep->right = a;\
       ep->left = NIL;\
       ep->content = CNIL;\
       latexout("d");\
     }

#define _DERREV \
     { env_t *tmp = ep->right;\
       ep->right = tmp -> right;\
       if (ep->right != NIL) ep->right->prev = ep;\
       ep->left = tmp -> left;\
       if (ep->left != NIL) ep->left->prev = ep;\
       ep->content = tmp->content;\
       h_free(henv,tmp);\
       latexout("d*");\
     };

#define _QUESTION \
       ep->prev->right = ep->left;\
       if (ep->left != NIL)\
          ep->left->prev = ep->prev;\
       ep->prev = tmp->prev;\
       if (ep->prev != NIL) \
          ep->prev->right = ep;\
       else\
          context = ep;\
       ep->left = tmp;\
       tmp->prev = ep;\
       latexout("t");

#define _QUESTIONREV \
       if (ep->prev != NIL)\
          ep->prev->right = ep->left;\
       else\
          context = ep->left;\
       ep->left->prev = ep->prev;\
       if (tmp->right != NIL)\
          tmp->right->prev = ep;\
       ep->left = tmp->right;\
       tmp->right = ep;\
       ep->prev = tmp;\
       latexout("t*");

#define _CONT(val) \
     { cont_t *tmp;\
       h_malloc(hcont,tmp);\
       tmp->d = val;\
       tmp->next = ep->content;\
       ep->content = tmp;\
     }

#define _CONTRET				\
     { cont_t *tmp = ep->content;		\
       void *add = tmp->d;			\
       ep->content = tmp->next;			\
       h_free(hcont,tmp);		\
       latexout("r*");				\
       goto *add;				\
     }

void copydatac()
{
  stack_t tmp = dstack;
  
  dstack2.nbr = NBLK-1;
  dstack2.lastblk = SNIL;
  while (! s_empty(tmp)) {
    s_push(dstack2, s_top(tmp));
    s_next(tmp);
  }

  h_malloc(henv,ep);
  ep->left = ep->right = ep->prev = NIL;
  ep->content = CNIL;
  context = ep;
};
