/* Macro include file for the call by value machine */

/* Possible contents of ep->l */
#define _L 1
#define _v 2  /* Value */

#define NIL (env_t *) NULL

typedef union {
        void *addr;       /* if l = _L or 0 */
        signed long val;  /* if l = _v */
        /* nothing */     /* if l = _h, _l or _i */
} v_t;
 
typedef struct env_s {
        struct env_s *prev, *left, *right;
        v_t v;
        bool l;
} env_t;

heap_t henv = h_init(env_t);
stack_t dstack = s_init;
env_t *ep,*context;

#ifdef latex

void printenv(env_t *env)
{
  if (env == NIL) 
     fprintf(f,"\\carre");
  else {
     if (env == ep)
        fprintf(f,"{\\it ep:}");
     fprintf(f, "$<$");
     switch (env->l) {
       case _v : fprintf(f,"%li",env->v.val);
                 break;
       case _h : fprintf(f, "h");
                 break; 
       case _l : fprintf(f, "l");
                 break; 
       case _i : fprintf(f, "i");
                 break; 
       case _L : printenv(env->left);
                 fprintf(f,":L");
                 break;
       case 0 :printenv(env->left);
               break;
       default : fprintf(f,"Bad");
       }
     fprintf(f, ",");
     printenv(env->right);
     fprintf(f, "$>$");
   }
};

char *latexpreamble()
{
   return "\\tvi\\quad#\\hfill\\quad&\n\
\\tvi\\hfill\\quad#\\hfill\\quad\\tvi\\cr\n\
\\noalign{\\hrule}\n\
Instruc.&   Environment\\cr\n";
};

void latexprintline(char *s)
{
  fprintf(f,"{\\it %s}& ",s);
  printenv(context);
  fprintf(f, "\\cr\n");
};

#endif

#define _INIT(lbl) \
     { initstack(&&lbl);\
       context = ep;\
       latexout("beginning");\
     }

#define _DER(l1out) \
     { env_t *a;\
       h_malloc(henv,a);\
       a->right = ep->right;\
       if (a->right != NIL) a->right->prev = a;\
       a->left = ep->left;\
       a->v = ep->v;\
       a->l = ep-> l;\
       a->prev = ep;\
       ep->l = _L;\
       ep->right = a;\
       ep->left = copyenv();\
       ep->v.addr = &&l1out;\
     }

#define _DERNS(l1out) \
     { env_t *a;\
       h_malloc(henv,a);\
       a->right = ep->right;\
       if (a->right != NIL) a->right->prev = a;\
       a->left = ep->left;\
       a->v = ep->v;\
       a->l = ep-> l;\
       a->prev = ep;\
       ep->l = _L;\
       ep->right = a;\
       ep->left = NIL;\
       ep->v.addr = &&l1out;\
     }

#define _DERREV \
     { env_t *tmp = ep->right;\
       freetree(ep->left);\
       ep->right = tmp -> right;\
       if (ep->right != NIL) ep->right->prev = ep;\
       ep->left = tmp->left;\
       ep->v = tmp->v;\
       ep->l = tmp->l;\
       h_free(henv,tmp);\
     }

#define _PUTVAL(x,mess) \
    { void *add = ep->v.addr;\
      putenv(ep->left);\
      ep->left = NIL;\
      ep->l = _v;\
      ep->v.val = x;\
      ep = ep->right;\
      latexout(mess);\
      goto *add;\
    }

#define _POPCONTEXTVALUE \
    { env_t *tmp = ep->prev;\
      tmp->right = ep->right;\
      tmp->left = ep->left;\
      /* The left branch of ep->prev doesn't contain any environment */\
      /* because ep->prev must be a 'value environment'              */\
      /* ie l = _v, left = NIL, v.val = value                        */\
      tmp->v = ep->v;\
      tmp->l = ep->l;\
      if (tmp->right != NIL)\
         tmp->right->prev = tmp;\
      h_free(henv,ep);\
      ep = tmp;\
    }
    
#define _SWAPCONT \
    { env_t *tmpcont = ep->left;\
\
      if (ep->prev != NIL) {\
            ep->left = context;\
            ep->prev->right = NIL;\
      } else \
            ep->left = NIL;\
      if (tmpcont != NIL) {\
            context = tmpcont;\
            while (tmpcont->right != NIL)\
               tmpcont = tmpcont->right;\
            tmpcont->right = ep;\
            ep->prev = tmpcont;\
      } else {\
            context = ep;\
            ep->prev = NIL;\
      }\
    }

env_t *copytree(env_t *p, env_t *pr)
{
  if (p != NIL) {
    env_t *a;
    h_malloc(henv,a);
#ifdef count
    nbr[2]++;
#endif
    a->left = copytree(p->left, NIL);
    a->right = copytree(p->right,a);
    a->prev = pr;
    a->v = p->v;
    a->l = p->l;
    return a;
  }
  else return NIL;
};


env_t *copyenv()
{
  env_t *res;

  if (ep->prev != NIL) {
    ep->prev->right = NIL; /* Disconnect the stack from the environment */
    res = copytree(context, NIL);
    ep->prev->right = ep; /* Reconnect stack and environment */
    return res;
  }
  else return NIL;
};

void freetree(env_t *p)
{
  if (p != NIL) {
    freetree(p->left);
    freetree(p->right);
    h_free(henv,p);
  }
};

void putenv(env_t *p)
{
  if (ep->prev != NIL) {
    ep->prev->right = NIL; /* disconnect stack/environment */
    freetree(context);
  }
  if (p != NIL) {
    context = p;
    while (p->right != NIL)
       p = p->right;
    p->right = ep;
    ep->prev = p;
  }
  else {
    context = ep;
    ep->prev = NIL;
  }
};

void initstack(void *add)
{
   env_t *tmp;
   stack_t dstacktmp = dstack;
   ulng i;
   
   h_malloc(henv,ep);
   ep->v.addr = add;
   ep->left = NIL;
   ep->l = 0;
   h_malloc(henv,ep->right);\
   ep->right->right = ep->right->left = NIL;\
   ep->right->v.addr = NULL;\
   ep->right->l = 0;\
   ep->right->prev = ep;\
   while (! s_empty(dstacktmp)) {
      h_malloc(henv,tmp);
      tmp->right = ep;
      ep->prev = tmp;
      tmp->left = NIL;
      tmp->l = ((s_top(dstacktmp) & 1) == 0 ? _h : _i);
      ep = tmp;
      for (i = 1; i<=s_top(dstacktmp)/2; i++) {
         h_malloc(henv,tmp);
         tmp->right = ep;
         ep->prev = tmp;
         tmp->left = NIL;
         tmp->l = _l;
         ep = tmp;
      }
      s_next(dstacktmp);
   }
   ep->prev = NIL;
};
