/* Macro include file for the GOI machine (non optimized) */

#define NIL (env_t *) NULL
#define CNIL (cont_t *) NULL
#define DNIL (dat_ac_t *)NULL
     
typedef struct cont {
        unsigned char d;
        struct cont *next;
} cont_t;
 
typedef struct env {
        struct env *prev, *right, *left;
        cont_t *content;
} env_t;

typedef struct dat_ac_s {
        struct dat_ac_s *next;
        unsigned char d;
} dat_ac_t;

ulng R0 = 1;
stack_t vstack = s_init;
heap_t henv = h_init(env_t);
heap_t hcont = h_init(cont_t);
heap_t hdat = h_init(dat_ac_t);
dat_ac_t *dstack = DNIL;
env_t tree1 = {NIL,NIL,NIL,CNIL};
env_t tree2 = {NIL,NIL,NIL,CNIL};
env_t tree3 = {NIL,&tree1,&tree2,CNIL};
env_t *ep = &tree3,*context = &tree3;

#ifdef latex

void printenv(env_t *env)
{
  cont_t *a = env->content;

  while (a != CNIL)
  {
     if (a->d) 
        fprintf(f, "R:");
     else
        fprintf(f, "L:");
     a = a->next;
   }
  if (env == ep) 
     fprintf(f,"{\\it ep:}");
  if ((env->right == NIL) & (env->left == NIL))
     fprintf(f,"\\carre");
  else {
     fprintf(f, "$<$");
     printenv(env->left);
     fprintf(f, ",");
     printenv(env->right);
     fprintf(f, "$>$");
   }
};

char *latexpreamble()
{
  return "\\tvi\\quad#\\hfill\\quad&\n\
\\tvi\\hfill\\quad#\\quad&\n\
\\tvi\\hfill\\quad#\\hfill\\quad&\n\
\\tvi\\hfill\\quad#\\quad&\n\
\\tvi\\hfill\\quad#\\quad\\tvi\\cr\n\
\\noalign{\\hrule}\n\
Instruction&  Mult&   Environment& Data access& Data\\cr\n";
};

void latexprintline(char *s)
{
  ulng a = R0;
  stack_t tmp = vstack;
  dat_ac_t *d = dstack;

  fprintf(f, "{\\it %s}& {\\it ",s);
  while (a > 1)
  {
     if (a & 1) 
        fprintf(f, "r:");
     else
        fprintf(f, "l:");
     a >>= 1;
   };
  fprintf(f, "}\\carre & ");
  printenv(context);
  fprintf(f, "& ");
  while (d != DNIL)
  {
    fprintf(f, "%c:", d->d);
    d = d->next;
  }
  fprintf(f, "\\carre & ");  
  while (! s_empty(tmp))
  {
    fprintf(f, "%li:", s_top(tmp));
    s_next(tmp);
  }
  fprintf(f, "\\carre\\cr\n");
};

#endif

#define _DER \
     { env_t *a, *b;\
       h_malloc(henv,a);\
       h_malloc(henv,b);\
       a->right = ep->right;\
       if (a->right != NIL) a->right->prev = a;\
       a->left = ep->left;\
       if (a->left != NIL) a->left->prev = a;\
       a->content = ep->content;\
       a->prev = ep;\
       ep->right = a;\
       ep->left = b;\
       b->prev = ep;\
       b->right = b->left = NIL;\
       b->content = ep->content = CNIL;\
       latexout("d");\
     }

#define _DERREV \
     { env_t *tmp = ep->right;\
       h_free(henv, ep->left);\
       ep->right = tmp -> right;\
       if (ep->right != NIL) ep->right->prev = ep;\
       ep->left = tmp -> left;\
       if (ep->left != NIL) ep->left->prev = ep;\
       ep->content = tmp->content;\
       h_free(henv,tmp);\
       latexout("d*");\
     };

#define _QUESTION \
     { env_t *tmp = ep;\
       cont_t *b = ep->prev->content;\
       ep = ep->prev;\
       ep->right = tmp->right;\
       if (ep->right != NIL) ep->right->prev = ep;\
       tmp->right = tmp->left;\
       tmp->left = ep->left;\
       if (tmp->left != NIL) tmp->left->prev = tmp;\
       ep->left = tmp;\
       tmp->prev = ep;\
       ep->content = tmp->content;\
       tmp->content = b;\
       latexout("t");\
     };

#define _QUESTIONREV \
     { env_t *tmp = ep->left;\
       cont_t *b = ep->content;\
       ep->left = tmp->left;\
       if (ep->left != NIL) ep->left->prev = ep;\
       tmp->left = tmp->right;\
       tmp->right = ep->right;\
       if (tmp->right != NIL) tmp->right->prev = tmp;\
       ep->right = tmp;\
       tmp->prev = ep;\
       ep->content = tmp->content;\
       tmp->content = b;\
       ep = ep->right;\
       latexout("t*");\
     };

#define _DSTACKI \
          /* Add 'i' at the end of dstack */\
          {  dat_ac_t **tmp = &dstack, *tmp2;\
\
             while (*tmp != DNIL)\
                tmp = &((*tmp)->next);\
             h_malloc(hdat,tmp2);\
             tmp2->next = DNIL;\
             tmp2->d = 'i';\
             *tmp = tmp2;\
          }

#define _DSTACKH \
          /* Replace i by h  at the end, and print ; if we are not at */\
          /* the beginning of a list */\
          if (dstack->next != DNIL) {\
             dat_ac_t *tmp2 = dstack;\
             while (tmp2->next->next != DNIL)\
                tmp2 = tmp2->next;\
             if (tmp2->d != 'h') \
                printf("; ");\
             tmp2->next->d = 'h';\
          } else\
             dstack->d = 'h';
             
#define _DSTACKL \
          /* Replace h by l at the end of dstack */\
          {  dat_ac_t *tmp2 = dstack;\
             while (tmp2->next != DNIL)\
                tmp2 = tmp2->next;\
             tmp2->d = 'l';\
          }

#define _CUTDSTACK \
     /* Cut dstack after the last 'h' */\
     {  dat_ac_t **tmp = &dstack, *tmp2 = dstack;\
        while (tmp2->next != DNIL) {\
           if (tmp2->d == 'h')\
              tmp = &tmp2->next;\
           tmp2 = tmp2->next;\
        } \
        freedatac(*tmp);\
        *tmp = DNIL;\
     }

void freedatac(dat_ac_t *p)
{
  if (p != DNIL) {
    freedatac(p->next);
    h_free(hdat,p);
  };
};
