/* Include file for all SFL machines */

#include <sys/stdtypes.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <malloc.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/times.h>

/* Define CLOCKS_PER_SEC */

#ifndef CLOCKS_PER_SEC
#   ifdef CLK_TCK
#      define CLOCKS_PER_SEC CLK_TCK
#   else
#      define CLOCKS_PER_SEC 60
#   endif
#endif

/* Number of lines in a page */

#define LINEMAX 40

/* Number of elements in a memory block */

#define NBLK 100

#define tt 1
#define ff 0
#define _h 3
#define _l 4
#define _i 5

typedef unsigned long ulng;
typedef unsigned char bool;

/* Putting limits */

void putlimit(int resource, int limit)
{
   struct rlimit maximum = {limit, limit};
   
   if (setrlimit(resource, maximum) != 0)
      printf("Warning : limit setting failed\n");
};


#ifdef count

ulng nbr[26], maxmem, actmem = 0, totmem;

#endif

/* Data structures : this one accepts any free and malloc */

typedef struct helem_s
{
  struct helem_s *next_el;
} helem_t;

typedef struct hblk_s
{
  struct hblk_s *next_blk;
  char data[0];
} hblk_t;

typedef struct heap_s
{
  unsigned short size;
  helem_t *pfree;
  hblk_t *palloc;
} heap_t;

#define ENIL (helem_t *)NULL
#define BNIL (hblk_t *)NULL
#define h_init(t) {sizeof(t), ENIL, BNIL}

#define tmpelem(n) ((helem_t *)(&tmp->data[(n)*h->size]))

void blk_malloc(heap_t *h, helem_t **x)
     {
        short i;
        hblk_t *tmp = (hblk_t *)malloc(sizeof(hblk_t)+NBLK * h->size);
        tmp->next_blk = h->palloc;
        h->palloc = tmp;
        tmpelem(NBLK-1)->next_el = h->pfree;
        h->pfree = tmpelem(1);
        for (i=1; i<NBLK-1; i++)
           tmpelem(i)->next_el = tmpelem(i+1); 
        *x = tmpelem(0);
      }
      
#ifdef count

void hfree(heap_t *h, helem_t *p)
{
   actmem-=h->size;
   p->next_el=h->pfree;
   h->pfree=p;
 };

#define h_free(h,p) hfree(&h, (helem_t *)p)

void hmalloc(heap_t *h, helem_t **x)
{
   actmem+=h->size;
   totmem+=h->size;
   if (actmem > maxmem)
      maxmem = actmem;
   if (h->pfree != ENIL)
     {
        *x = h->pfree;
        h->pfree = h->pfree->next_el;
      }
   else
     blk_malloc(h,x);
 };
 
#define h_malloc(h,x) hmalloc(&h, (helem_t **)(&x))

void hfreeall(heap_t *h)
{
   hblk_t *tmp;

   while (h->pfree != ENIL)
     {
        h->pfree = h->pfree->next_el;
        actmem += h->size;
      }
   while (h->palloc != BNIL)
     {
        tmp = h->palloc;
        h->palloc = h->palloc->next_blk;
        free(tmp);
        actmem -= NBLK * h->size;
     }
 };

#define h_freeall(h) hfreeall(&h)

#else

#define h_free(h,p) \
{\
   ((helem_t *) p)->next_el=h.pfree;\
   h.pfree=(helem_t *)p;\
}

#define h_malloc(h,x) \
{\
   if (h.pfree != ENIL) {\
        (helem_t *)x = h.pfree;\
        h.pfree = h.pfree->next_el;\
   } else blk_malloc(&h, (helem_t **)(&x));\
}

#define h_freeall(h) \
{\
   h.pfree = ENIL;\
   while (h.palloc != BNIL)\
     {\
        hblk_t *tmp = h.palloc;\
        h.palloc = h.palloc->next_blk;\
        free(tmp);\
     }\
}

#endif

/* Data structures : stack */

typedef struct stblk_s
{ 
   struct stblk_s *snext_blk;
   signed long v[NBLK];
} stblk_t;

typedef struct stack_s
{
   short nbr;
   stblk_t *lastblk;
} stack_t;

#define SNIL (stblk_t *) NULL
#define s_init {NBLK-1, SNIL}
#define s_top(s) s.lastblk->v[s.nbr]
#define s_empty(s) ((s.nbr == NBLK-1) && (s.lastblk == SNIL))

#ifdef count

void pop(stack_t *s)
{
   actmem -= sizeof(signed long);
   if (s->nbr == 0)
   {
      stblk_t *vtmp = s->lastblk;
      s->lastblk = s->lastblk->snext_blk;
      free(vtmp);
      s->nbr = NBLK;
   }
   s->nbr--;
};

#define s_pop(s) pop(&s)

void push(stack_t *s, signed long val)
{
   actmem += sizeof(signed long);
   totmem += sizeof(signed long);
   if (actmem > maxmem)
      maxmem = actmem;
   s->nbr++;
   if (s->nbr >= NBLK)
   {
      stblk_t *vtmp = (stblk_t *)malloc(sizeof(stblk_t));
      s->nbr = 0;
      vtmp->snext_blk = s->lastblk;
      s->lastblk = vtmp;
   }
   s->lastblk->v[s->nbr] = val;
};

#define s_push(s,val) push(&s,val)

void snext(stack_t *s)
{
   if (s->nbr == 0)
   {
      s->lastblk = s->lastblk->snext_blk;
      s->nbr = NBLK;
   }
   s->nbr--;
};

#define s_next(s) snext(&s)
  
#else

#define s_pop(s) \
{\
   if (s.nbr == 0)\
   {\
      stblk_t *vtmp = s.lastblk;\
      s.lastblk = s.lastblk->snext_blk;\
      free(vtmp);\
      s.nbr = NBLK;\
   }\
   s.nbr--;\
}

#define s_push(s, val) \
{\
   s.nbr++;\
   if (s.nbr >= NBLK)\
   {\
      stblk_t *vtmp = (stblk_t *)malloc(sizeof(stblk_t));\
      s.nbr = 0;\
      vtmp->snext_blk = s.lastblk;\
      s.lastblk = vtmp;\
   }\
   s.lastblk->v[s.nbr] = val;\
}

#define s_next(s) \
{\
   if (s.nbr == 0)\
   {\
      s.lastblk = s.lastblk->snext_blk;\
      s.nbr = NBLK;\
   }\
   s.nbr--;\
}

#endif    

/* Chrono functions */

struct tms timebufstart,timebufend;

ulng ticktoms(clock_t tk)
{
  return (tk / CLOCKS_PER_SEC * 1000 + ((tk % CLOCKS_PER_SEC) * 1000 + 
           (CLOCKS_PER_SEC - 1) / 2) / CLOCKS_PER_SEC);
};  

/* Latex trace functions */

#ifdef latex

FILE *f;

int linesinpage;

void latexstartall()
{
  f = fopen("a.tex", "w");
  fprintf(f, "\\documentstyle[a4]{article}\n\n\
\\def\\tvi{\\vrule height 10pt depth 5pt }\n\
\\def\\carre{\\unskip\\kern 1pt\\penalty 500\n\
\\raise -1pt\\hbox{\\vrule\\vbox to6pt{\\hrule width 6pt\n\
\\vfill\\hrule}\\vrule}\\kern 1pt}\n\
\\begin{document}\n\n");
  linesinpage = 0;
};

void latexendall()
{
  fprintf(f,"\\end{document}");
  fclose(f);
};

char *latexpreamble();

void latexprintline(char *s);

#else

#define latexstartall() {}
#define latexendall() {}

#endif

/* Latex output and/or count routines */

void latexstart()
{
#ifdef count
  unsigned char i;

  for (i=0; i<26; i++)
     nbr[i]=0;
  maxmem = actmem;
  totmem = 0;
#endif

  times(&timebufstart);

#ifdef latex
  fprintf(f,"\\vbox{\\offinterlineskip\\halign{");
  fprintf(f,latexpreamble());
  fprintf(f,"\\noalign{\\hrule}\n");
  linesinpage++;
#endif

};

void latexend()
{
#ifdef count
  unsigned char i;

  if (nbr[0])
     printf("Number of ! and !* : %lu\n",nbr[0]);
  for (i=3; i<26; i++)
     if (nbr[i])
        printf("Number of %c and %c* : %lu\n",i+'a',i+'a',nbr[i]);
  if (nbr[1])
     printf("Number of printings, arithmetic and other operations : %lu\n",nbr[1]);
  if (nbr[2])
     printf("Number of copied cells : %lu\n",nbr[2]);
  printf("Maximum memory requirement : %lu bytes.\n",maxmem);
  printf("Total memory requirement : %lu bytes.\n",totmem);
  if (actmem>0)
     printf("Kept memory space : %lu bytes.\n",actmem);
#endif

#if ((defined count) || (defined latex))
  printf("These times are for information only\n");
#endif
  times(&timebufend);
  printf("User time : %lu ms\n", 
     ticktoms(timebufend.tms_utime-timebufstart.tms_utime));
  printf("System time : %lu ms\n",
     ticktoms(timebufend.tms_stime-timebufstart.tms_stime));

#ifdef latex
  fprintf(f, "\\noalign{\\hrule}\n}}\n\\kern 1cm\n");
  linesinpage += 2;
  if (linesinpage > LINEMAX-1)
  {
     fprintf(f, "\\pagebreak\n");
     linesinpage = 0;
  }
#endif

};

#if ((defined count) || (defined latex))

void latexout(char *s)
{
#ifdef latex
  if (linesinpage > LINEMAX)
  {
     fprintf(f,"}}\n\\pagebreak\n\\vbox{\\offinterlineskip\\halign{");
     fprintf(f, latexpreamble());
     fprintf(f,"\\noalign{\\hrule}\n");
     linesinpage = 1;
  }
  latexprintline(s); 
  linesinpage++;
  fflush(f);
#endif

#ifdef count
  if (*s == '!')
     nbr[0]++;
  else if ((*s>'c') && (*s<='z') && ((s[1] == 0) || (s[1] == '*'))) 
     nbr[*s-'a']++;
  else 
     nbr[1]++;
#endif
};


#else

#define latexout(x) {}

#endif
